"""
The MIT License (MIT)

Copyright (c) 2021-present UnrealFar & TheGenocides
Copyright (c) 2023-present Sengolda

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""

from __future__ import annotations

import base64
import random
import string
import datetime

from random import randint
from typing import Tuple, Optional, Literal, TYPE_CHECKING
from requests_oauthlib import OAuth1
from .errors import pytwotException

if TYPE_CHECKING:
    from .http import HTTPClient

__all__ = ("OauthSession", "Scope")


class Scope:
    """Scopes allow you to set granular access for your App so that your App only has the permissions that it needs. Here are the full documented scopes!

    .. list-table:: Scopes
        :widths: 33, 33
        :header-rows: 1

        *   - Scope
            - Description
        *   - tweet.read
            - All the Tweets you can view, including Tweets from protected accounts.
        *   - tweet.write
            - Tweet and Retweet for you.
        *   - tweet.moderate.write
            - Hide and unhide replies to your Tweets.
        *   - users.read
            - Any account you can view, including protected accounts.
        *   - follows.read
            - People who follow you and people who you follow.
        *   - follows.write
            - Follow and unfollow people for you.
        *   - offline.access
            - Stay connected to your account until you revoke access.
        *   - space.read
            - All the Spaces you can view.
        *   - mute.read
            - Accounts you’ve muted.
        *   - mute.write
            - Mute and unmute accounts for you.
        *   - like.read
            - Tweets you’ve liked and likes you can view.
        *   - like.write
            - Like and un-like Tweets for you.
        *   - list.read
            - Lists, list members, and list followers of lists you’ve created or are a member of, including private lists..
        *   - list.write
            - Create and manage Lists for you..
        *   - block.read
            - Accounts you’ve blocked.
        *   - block.write
            - Block and unblock accounts for you.
        *   - bookmarks.read
            - Get Bookmarked Tweets from an authenticated user.
        *   - bookmarks.write
            - Bookmark and remove Bookmarks from Tweets.


    .. versionadded:: 1.5.0
    """

    def __init__(
        self,
        *,
        tweet_read: bool = False,
        tweet_write: bool = False,
        tweet_moderate_write: bool = False,
        users_read: bool = False,
        follows_read: bool = False,
        follows_write: bool = False,
        offline_access: bool = False,
        space_read: bool = False,
        mute_read: bool = False,
        mute_write: bool = False,
        like_read: bool = False,
        like_write: bool = False,
        list_read: bool = False,
        list_write: bool = False,
        block_read: bool = False,
        block_write: bool = False,
        bookmarks_read: bool = False,
        bookmarks_write: bool = False,
    ):
        self.tweet_read = tweet_read  # "tweet.read" if tweet_read else None
        self.tweet_write = tweet_write  # "tweet.write" if tweet_write else None
        self.tweet_moderate_write = tweet_moderate_write  # "tweet.moderate.write" if tweet_moderate_write else None
        self.users_read = users_read  # "users.read" if users_read else None
        self.follows_read = follows_read  # "follows.read" if follows_read else None
        self.follows_write = follows_write  # "follows.write" if follows_write else None
        self.offline_access = offline_access  # "offline.access" if offline_access else None
        self.space_read = space_read  # "space.read" if space_read else None
        self.mute_read = mute_read  # "mute.read" if mute_read else None
        self.mute_write = mute_write  # "mute.write" if mute_write else None
        self.like_read = like_read  # "like.read" if like_read else None
        self.like_write = like_write  # "like.write" if like_write else None
        self.list_read = list_read  # "list.read" if list_read else None
        self.list_write = list_write  # "list.write" if list_write else None
        self.block_read = block_read  # "block.read" if block_read else None
        self.block_write = block_write  # "block.write" if block_write else None
        self.bookmarks_read = bookmarks_read  # "bookmarks.read" if bookmarks_read else None
        self.bookmarks_write = bookmarks_write  # "bookmarks.write" if bookmarks_write else None

    def __repr__(self) -> str:
        s = "Scopes("
        for attr in dir(self):
            if "_read" in attr or "_write" in attr or "_access" in attr:
                scope = getattr(self, attr, None)
                s += f"{attr}={scope} "
        return s.rstrip(" ") + ")"

    @classmethod
    def read_only(cls, *, offline_access: bool = False):
        """A classmethod that enables only read scopes. offline_access scope is optional, you can set it true or not. Defaults to False.

        Parameters
        ------------
        offline_access: :class:`bool`
            Indicates to stay connected to your account until you revoke access.


        .. versionadded:: 1.5.0
        """
        return cls(
            tweet_read=True,
            users_read=True,
            follows_read=True,
            space_read=True,
            mute_read=True,
            like_read=True,
            list_read=True,
            block_read=True,
            bookmarks_read=True,
            offline_access=offline_access,
        )

    @classmethod
    def write_only(cls, *, offline_access: bool = False):
        """A classmethod that enables only write scopes. This includes tweet_moderate_write scope. offline_access scope is optional, you can set it true or not. Defaults to False.

        Parameters
        ------------
        offline_access: :class:`bool`
            Indicates to stay connected to your account until you revoke access.


        .. versionadded:: 1.5.0
        """
        return cls(
            tweet_write=True,
            follows_write=True,
            mute_write=True,
            like_write=True,
            list_write=True,
            block_write=True,
            tweet_moderate_write=True,
            bookmarks_write=True,
            offline_access=offline_access,
        )

    @classmethod
    def all(cls, *, offline_access: bool = False):
        """A classmethod that enables all scopes. offline_access scope is optional, you can set it true or not. Defaults to False.

        Parameters
        ------------
        offline_access: :class:`bool`
            Indicates to stay connected to your account until you revoke access.


        .. versionadded:: 1.5.0
        """
        return cls(
            tweet_read=True,
            tweet_write=True,
            tweet_moderate_write=True,
            users_read=True,
            follows_read=True,
            follows_write=True,
            offline_access=offline_access,
            space_read=True,
            mute_read=True,
            mute_write=True,
            like_read=True,
            like_write=True,
            list_read=True,
            list_write=True,
            block_read=True,
            block_write=True,
            bookmarks_read=True,
            bookmarks_write=True,
        )

    @property
    def value(self) -> str:
        """:class:`str`: Returns the scope url encoded value. Example:

        .. code-block:: py

            import pytwot

            scope = pytwot.Scope(tweet_read=True, tweet_write=True)
            print(scope.value) # --> 'tweet.read%20tweet.write'


        .. versionadded:: 1.5.0
        """
        value = ""
        for attr in dir(self):
            if "_read" in attr or "_write" in attr or "_access" in attr:
                scope = getattr(self, attr, None)
                if scope:
                    value += f"{attr.replace('_', '.')}%20"

        return value.rstrip("%20")  # Only remove the last %20.


class OauthSession:
    """Represents an OauthSession for OAuth1 and OAuth2 Authorization.

    Parameters
    ------------
    consumer_key: Optional[:class:`str`]
        The application's consumer key.
    consumer_secret: Optional[:class:`str`]
        The application's consumer secret.
    access_token: Optional[:class:`str`]
        The application's access token.
    access_token_secret: Optional[:class:`str`]
        The application's access token secret.
    http_client: :class:`HTTPClient
        The :class:`HTTPClient` for making requests.
    callback_url: Optional[:class:`str`]
        The callback url, the user will get redirect to the callback url after they authorize. Default to None.
    client_id: Optional[:class:`str`]
        The client's OAuth 2.0 Client ID from keys and tokens page.
    client_secret: Optional[:class:`str`]
        The client's OAuth 2.0 Client Secret from keys and tokens page.


    .. versionadded:: 1.2.0
    """

    def __init__(
        self,
        consumer_key: Optional[str],
        consumer_secret: Optional[str],
        *,
        access_token: Optional[str],
        access_token_secret: Optional[str],
        http_client: HTTPClient,
        callback_url: Optional[str] = None,
        client_id: Optional[str] = None,
        client_secret: Optional[str] = None,
    ) -> None:
        self.http_client: HTTPClient = http_client
        self.consumer_key = consumer_key
        self.consumer_secret = consumer_secret
        self.access_token = access_token
        self.access_token_secret = access_token_secret
        self.callback_url = callback_url
        self.client_id = client_id
        self.client_secret = client_secret

    @property
    def oauth1(self) -> OAuth1:
        """:class:`Oauth1`: Wrap the credentials in a function that return Oauth1. Usually Uses for Authorization.

        .. versionadded:: 1.2.0
        """
        return OAuth1(
            self.consumer_key,
            client_secret=self.consumer_secret,
            resource_owner_key=self.access_token,
            resource_owner_secret=self.access_token_secret,
            callback_uri=self.callback_url,
            decoding=None,
        )

    @property
    def basic_auth(self) -> str:
        """:class:`str`: The decoded base64 encoded client id and secret.

        .. versionadded:: 1.5.0
        """
        if not self.client_id and not self.client_secret:
            raise pytwotException("'client_id' and 'client_secret' argument is missing in your client instance!")

        return base64.b64encode(bytes(f"{self.client_id}:{self.client_secret}", "utf-8")).decode()

    def invalidate_access_token(self) -> None:
        """Invalidate the access token and access token secret of yout client.

        .. warning::
            This staticmethod will invalidate your access token and access token secret of your client.


        .. versionadded:: 1.3.5
        """
        self.http_client.request("POST", "1.1", "/oauth/invalidate_token", auth=True)

    def verify_credentials(self, *, raise_error: bool = True) -> Optional[bool]:
        """Verify the credentials are correct. Returns a boolean whether its succesful or not if raise_error turns to False. Default to True

        Parameters
        ------------
        raise_error: :class:`bool`
            Indicates whether to raise if the credentials are wrong. If sets to False, the method will returns a boolean, True for succesful and False for error.

        Raises
        --------
        :class:`Forbidden`
            Raised if the credentials are wrong.


        .. versionadded:: 1.5.0
        """
        try:
            error = None
            self.http_client.request("GET", "1.1", "/account/verify_credentials.json", auth=True)
        except Exception as e:
            error = e
        finally:
            if not error:
                return True
            elif error and raise_error:
                raise error
            else:
                return False

    def generate_request_tokens(
        self, access_type: Optional[Literal["read", "write", "direct_messages"]] = None
    ) -> dict:
        """Generates request tokens with an access_type. This method returns a dictionary with the credentials, example return object:

        .. code-blocks: python

            data = oauth_session.generate_request_tokens()
            print(data) # --> {'oauth_token': 'XXXXXXXXXXXXXXX', 'oauth_token_secret': 'XXXXXXXXXXXXXXX', 'oauth_callback_confirmed': 'true'}


        Parameters
        ------------
        access_type: :class:`str`
            Must be either read, write, direct_messages. read for reading twitter info only, write will have the same as read and also write permission this includes but not limited to post & delete a :class:`Tweet`, and direct_messages is for read & write and sending & deleting

        Returns
        ---------
        :class:`dict`
            Returns the request_tokens data.


        .. versionadded:: 1.3.5
        """
        try:
            if access_type:
                url = f"/request_token?x_auth_access_type={access_type}"

            else:
                url = f"/request_token"

            request_tokens = self.http_client.request("POST", "oauth", url, auth=self.oauth1)
            data = {}
            for credential in request_tokens.split("&"):
                k, v = credential.split("=")
                data[k] = v
            return data
        except Exception as e:
            raise e

    def create_oauth_url(
        self,
        access_type: Optional[Literal["read", "write", "direct_messages"]] = None,
        *,
        signin_with_twitter: bool = False,
    ) -> Optional[str]:
        """Creates an oauth url with an access type. This is the 1st step of making a request on behalf of other users through oauth1.1 usercontext.

        The callback after pressing authorize button is your callback url that you passed in your :class:`Client`. The oauth_token and oauth_verifier will automatically appended in the callback url. If you are setting up a sign up button in your website to lookup the user's profile information, You have to setup a system where if the oauth_token or oauth_verifier is present in the url then, it will use :meth:`OauthSession.post_oauth_token` to post an oauth token and verifier to exchange with the user's access token and secret. If its for personal uses then just copy the result and passed in :meth:`OauthSession.post_oauth_token`.

        Parameters
        ------------
        access_type: :class:`str`
            Must be either read, write, direct_messages. read for reading twitter info only, write will have the same as read and also write permission this includes but not limited to post & delete a :class:`Tweet`, and direct_messages is for read & write and sending & deleting :class:`DirectMessages`.
        signin_with_twitter: :class:`bool`
           Register a user account in as little as one click. This works on websites, iOS, mobile, and desktop applications.

        Returns
        ---------
        :class:`str`
            Returns an oauth url.


        .. versionadded:: 1.3.5
        """
        if not self.callback_url:
            raise pytwotException("'callback_url' argument is missing in your client instance")

        access_type = access_type.lower()
        assert access_type in (
            "read",
            "write",
            "direct_messages",
        ), "Wrong access type passed! must be 'read', 'write', or 'direct_messages')"
        request_tokens = self.generate_request_tokens(access_type)
        authorize_url = (
            self.http_client.base_url + "oauth/authorize"
            if not signin_with_twitter
            else self.http_client.base_url + "oauth/authenticate"
        )
        return authorize_url + f"?oauth_token={request_tokens.get('oauth_token')}"

    def post_oauth_token(self, oauth_token: str, oauth_verifier: str) -> Optional[Tuple[str]]:
        """Posts the oauth token & verifier. This is the 2nd step(and the last step) of making a request on behalf of other users through oauth1.1 usercontext. Returns a pair of access token & secret also the user's username(present as screen_name) and id e.g ("access_token=xxxxxxxxxxxxx", "access_token_secret=xxxxxxxxxxxxx", "screen_name=TheGenocides", "user_id=1382006704171196419"). Uses the access token and secret to make request on behalf of users! You can use the raw api or construct another client with the access token and secret.

        Parameters
        ------------
        oauth_token: :class:`str`
            The Oauth token.
        oauth_verifier: :class:`str`
            The Oauth verifier.

        Returns
        ---------
        :class:`tuple`
            Returns a :class:`tuple` object with the credentials in.


        .. versionadded:: 1.3.5
        """
        res = self.http_client.request(
            "POST",
            "",
            "oauth/access_token",
            params={"oauth_token": oauth_token, "oauth_verifier": oauth_verifier},
        )

        return tuple(res.split("&"))

    def create_oauth2_url(
        self,
        scope: Scope,
        *,
        code_challenge_method: str = "plain",
        state: Optional[str] = None,
    ):
        """Creates an oauth 2 url. This is The 1st step of using OAuth 2.0 Authorization Code Flow with PKCE. The callback after pressing authorize button is your callback url that you passed in your :class:`Client`.

        Parameters
        ------------
        scope: :class:`Scope`
            The scope permissions. Must be an instance of :class:`Scope`.
        code_challenge_method: :class:`str`
            The code challenge method, must be tiher plain or s256. Default to plain.
        state: Optional[:class:`str`]
            A random string you provide to verify against CSRF attacks. If none specified, the method will generates one.


        .. versionadded:: 1.5.0
        """
        if not self.callback_url:
            raise pytwotException("'callback_url' argument is missing in your client instance")

        code_challenge_method = code_challenge_method.lower()
        assert code_challenge_method in [
            "plain",
            "s256",
        ], "Wrong code_challenge_method passed: must be 'plain' or 's256'"

        if not state:
            state = "".join(random.choices(string.ascii_uppercase + string.digits, k=10)).lower()
        else:
            state = base64.b64decode(state.encode())

        now = datetime.datetime.now()
        timestamp = now.strftime("%m%d%Y%H%M%S")
        random_append = randint(1000, 100000)
        rand_dec = randint(300, 800)
        state = base64.b64encode(state.encode()).decode()
        code_challenge = "{}{}.{}".format(timestamp, random_append, rand_dec)
        client_id = self.client_id
        response_type = "code"
        scope = scope.value

        return f"https://twitter.com/i/oauth2/authorize?response_type={response_type}&client_id={client_id}&redirect_uri={self.callback_url}&scope={scope}&state={state}&code_challenge={code_challenge}&code_challenge_method={code_challenge_method}"

    def post_auth_code(self, code: str, code_challenge: str):
        """Posts the authorize code and code challenge. This is The 2nd step of using OAuth 2.0 Authorization Code Flow with PKCE. In this method, the client will make a request and create a new bearer token. With this, you can make request on behalf of users.

        Parameters
        ------------
        code: :class:`str`
            The authorize code.
        code_challenge: :class:`str`
            The code challenge.


        .. versionadded:: 1.5.0
        """
        return self.http_client.request(
            "POST",
            "2",
            "/oauth2/token",
            data=f"code={code}&grant_type=authorization_code&client_id={self.client_id}&redirect_uri={self.callback_url}&code_verifier={code_challenge}",
            headers={
                "Authorization": f"Basic {self.basic_auth}",
                "Content-Type": "application/x-www-form-urlencoded",
            },
        )

    def request_new_token(self, refresh_token: str) -> Optional[dict]:
        """Request a new access token with the refresh token, to obtain a refresh token make sure to use :meth:`OauthSession.create_oauth2_url` with offline_access permission sets to True in the scope parameter.

        Parameters
        ------------
        refresh_token: :class:`str`
            Refresh tokens allow an application to obtain a new access token without prompting the user via the refresh token flow.

        Returns
        ---------
        Optional[:class:`dict`]
            This method returns a dict object with the new request token in that dict.


        .. versionadded:: 1.5.0
        """
        return self.http_client.request(
            "POST",
            "2",
            "/oauth2/token",
            data=f"refresh_token={refresh_token}&grant_type=refresh_token&client_id={self.client_id}",
            headers={
                "Authorization": f"Basic {self.basic_auth}",
                "Content-Type": "application/x-www-form-urlencoded",
            },
        )



"""
The MIT License (MIT)

Copyright (c) 2021-present UnrealFar & TheGenocides
Copyright (c) 2023-present Sengolda

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""

import datetime
from typing import Any, Union, Dict, List, Optional

from .enums import SpaceState
from .utils import time_parse_todt
from .user import User
from .tweet import Tweet
from .objects import Comparable
from .dataclass import Topic
from .constants import (
    TWEET_EXPANSION,
    USER_FIELD,
    MEDIA_FIELD,
    PLACE_FIELD,
    POLL_FIELD,
    TWEET_FIELD,
)

__all__ = ("Space",)


class Space(Comparable):
    """Represents a twitter space.

    .. versionadded:: 1.3.5
    """

    __slots__ = ("__original_payload", "_payload", "http_client", "_include")

    def __init__(self, data: Dict[str, Any], http_client: object):
        self.__original_payload = data
        self._includes = self.__original_payload.get("includes")
        self._payload = self.__original_payload.get("data") or self.__original_payload
        self.http_client = http_client
        super().__init__(self.id)

    def __repr__(self) -> str:
        return "Space(name={0.title} id={0.id} state={0.state})".format(self)

    @property
    def title(self) -> str:
        """:class:`str`: The space's title.

        .. versionadded:: 1.3.5
        """
        return self._payload.get("title")

    @property
    def id(self) -> str:
        """:class:`str`: The space's unique id.

        .. versionadded:: 1.3.5
        """
        return self._payload.get("id")

    @property
    def raw_state(self) -> str:
        """:class:`str`: The raw space's state in  a string.

        .. versionadded:: 1.3.5
        """
        return self._payload.get("state")

    @property
    def state(self) -> SpaceState:
        """:class:`SpaceState`: The type of the space's state.

        .. versionadded:: 1.3.5
        """
        return SpaceState(self.raw_state)

    @property
    def lang(self) -> str:
        """:class:`str`: The space's language.

        .. versionadded:: 1.3.5
        """
        return self._payload.get("lang")

    @property
    def created_at(self) -> datetime.datetime:
        """:class:`datetime.datetime`: Returns a datetime.datetime object with the space's created datetime.

        .. versionadded:: 1.3.5
        """
        return time_parse_todt(self._payload.get("created_at"))

    @property
    def started_at(self) -> Optional[datetime.datetime]:
        """Optional[:class:`datetime.datetime`]: Returns a datetime.datetime object with the space's started time. Only available if the space has started.

        .. versionadded:: 1.3.5
        """
        return time_parse_todt(self._payload.get("started_at")) if self._payload.get("started_at") else None

    @property
    def updated_at(self) -> Optional[datetime.datetime]:
        """Optional[:class:`datetime.datetime`]: Returns a datetime.datetime object with the space's last update to any of this Space's metadata, such as the title or scheduled time. Only available if the space has started.

        .. versionadded:: 1.3.5
        """
        return time_parse_todt(self._payload.get("updated_at"))

    @property
    def ticketed(self) -> bool:
        """Returns a bool indicate if the space is ticketed.

        Returns
        ---------
        :class:`bool`
            This method returns a :class:`bool` object.


        .. versionadded:: 1.5.0
        """
        return self._payload.get("is_ticketed")

    @property
    def topics(self) -> Optional[List[Topic]]:
        """Optional[List[:class:`Topic`]]: Returns a list of the space's topics, returns None if the space has no topic

        .. versionadded:: 1.5.0
        """
        if self._includes.get("topics"):
            return [Topic(**data) for data in self._includes["topics"]]
        return None

    @property
    def participant_count(self) -> int:
        """:class:`int`: Returns the current number of users in the Space, including Hosts and Speakers.

        .. versionadded:: 1.5.0
        """
        return int(self._payload.get("participant_count"))

    @property
    def subscriber_count(self) -> Optional[int]:
        """Optional[:class:`int`]: Returns the number of people who set a remainder to this Space. This requires you to authenticate the request using the Access Token of the creator of the requested Space aka using OAuth 2.0 Authorization Code with PKCE.

        .. versionadded:: 1.5.0
        """
        if self._payload.get("subscriber_count"):
            return int(self._payload["subscriber_count"])
        return None

    def fetch_creator(self) -> User:
        """Fetches the creator's using the id.

        Returns
        ---------
        :class:`User`
            This method returns a :class:`User` object.


        .. versionadded:: 1.3.5

        .. versionchanged:: 1.5.0
            Made as a function that returns :class:`User`.
        """
        id = int(self._payload.get("creator_id"))
        return self.http_client.fetch_user(id)

    def fetch_invited_users(self) -> Optional[List[User]]:
        """Fetches the invited users. Usually, users in this list are invited to speak via the Invite user option and have a Speaker role when the Space starts. Returns None if there isn't invited users.

        Returns
        ---------
        Optional[List[:class:`User`]]
            This method returns a list of users or an empty list if not found.


        .. versionadded:: 1.3.5

        .. versionchanged:: 1.5.0
            Made as a function that returns a list of :class:`User`.
        """
        if self._payload.get("invited_users"):
            ids = [int(id) for id in self._payload["invited_users"]]
            return self.http_client.fetch_users(ids)
        return None

    def fetch_hosts(self) -> Optional[List[User]]:
        """Fetches the space's hosts.

        Returns
        ---------
        Optional[List[:class:`User`]]
            Returns a list of :class:`User`.


        .. versionadded:: 1.3.5

        .. versionchanged:: 1.5.0
            Made as a function that returns a list of :class:`User`.
        """
        if self._payload.get("host_ids"):
            ids = [int(id) for id in self._payload["host_ids"]]
            return self.http_client.fetch_users(ids)
        return None

    def fetch_tweets(self) -> Optional[List[Tweet]]:
        """Fetches users who purchased a ticket to the space. This requires you to authenticate the request using the Access Token of the creator of the requested Space aka using OAuth 2.0 Authorization Code with PKCE.

        Returns
        ---------
        Optional[List[:class:`Tweet`]]
            This method returns a list of :class:`Tweet` objects or an empty list.


        .. versionadded:: 1.5.0
        """
        res = self.http_client.request(
            "GET",
            "2",
            f"/spaces/{self.id}/tweets",
            params={
                "expansions": TWEET_EXPANSION,
                "tweet.fields": TWEET_FIELD,
                "user.fields": USER_FIELD,
                "media.fields": MEDIA_FIELD,
                "place.fields": PLACE_FIELD,
                "poll.fields": POLL_FIELD,
            },
        )

        if not res or not res.get("data"):
            return []

        return [Tweet(data, http_client=self.http_client) for data in res["data"]]

    def fetch_buyers(self) -> Union[List[User], list]:
        """Fetches users who purchased a ticket to the space. This requires you to authenticate the request using the Access Token of the creator of the requested Space aka using OAuth 2.0 Authorization Code with PKCE.

        Returns
        ---------
        Union[List[:class:`User`], list]
            This method returns a list of users.


        .. versionadded:: 1.5.0
        """
        res = self.http_client.request(
            "GET",
            "2",
            f"/spaces/{self.id}/buyers",
            params={
                "expansions": TWEET_EXPANSION,
                "user.fields": USER_FIELD,
                "media.fields": MEDIA_FIELD,
                "place.fields": PLACE_FIELD,
                "poll.fields": POLL_FIELD,
                "tweet.fields": TWEET_FIELD,
            },
        )
        if not res:
            return []
        return [User(data, http_client=self.http_client) for data in res["data"]]

    def is_ticketed(self) -> bool:
        """An alias to :meth:`Space.ticketed`.

        Returns
        ---------
        :class:`bool`
            This method returns a :class:`bool` object.


        .. versionadded:: 1.3.5

        .. versionchanged:: 1.5.0
            Made as an alias to :meth:`Space.ticketed`.
        """
        return self.ticketed



"""
The MIT License (MIT)

Copyright (c) 2021-present UnrealFar & TheGenocides
Copyright (c) 2023-present Sengolda

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""

from typing import Any, Dict

from .enums import RelationsTypeEnum

__all__ = (
    "Relation",
    "RelationFollow",
    "RelationLike",
    "RelationRetweet",
    "RelationHide",
    "RelationUpdate",
    "RelationPin",
    "RelationDelete",
)


class Relation:
    """Represents a base class of all relations classes.

    .. versionadded:: 1.2.5
    """

    __slots__ = "_type"

    def __init__(self, type: int):
        self._type = RelationsTypeEnum(type)

    @property
    def type(self) -> RelationsTypeEnum:
        """:class:`RelationsTypeEnum`: Returns the relation type.

        .. versionadded:: 1.5.0
        """
        return self._type


class RelationFollow(Relation):
    """Represents a follow relation from a follow & unfollow request.

    .. versionadded:: 1.2.0
    """

    __slots__ = ("__original_payload", "_payload")

    def __init__(self, data: Dict[str, Any]):
        self.__original_payload = data
        self._payload = data.get("data")
        super().__init__(1 if self.following else 0)

    def __repr__(self) -> str:
        return "RelationFollow(type={0.type} following={0.following} pending={0.pending})".format(self)

    @property
    def pending(self) -> bool:
        """:class:`bool`: Check if the relation is pending.

        .. versionadded:: 1.2.0
        """
        return self._payload.get("pending", False)

    @property
    def following(self) -> bool:
        """:class:`bool`: Check if the relation is following.

        .. versionadded:: 1.2.0
        """
        return self._payload.get("following", False)


class RelationLike(Relation):
    """Represents a like relation from a like & unlike request.

    .. versionadded:: 1.2.0
    """

    __slots__ = ("__original_payload", "_payload")

    def __init__(self, data: Dict[str, Any]):
        self.__original_payload = data
        self._payload = data.get("data")
        super().__init__(2 if self.liked else None)

    def __repr__(self) -> str:
        return "RelationFollow(liked={0.liked})".format(self)

    @property
    def liked(self) -> bool:
        """:class:`bool`: Return True if user liked the tweet else False.

        .. versionadded:: 1.2.0
        """
        return self._payload.get("liked")


class RelationRetweet(Relation):
    """Represents a retweet relation from a retweet & unretweet request.

    .. versionadded:: 1.2.0
    """

    __slots__ = ("__original_payload", "_payload")

    def __init__(self, data: Dict[str, Any]):
        self.__original_payload = data
        self._payload = data.get("data")
        super().__init__(3 if self.retweeted else None)

    def __repr__(self) -> str:
        return "RelationRetweet(liked={0.retweeted})".format(self)

    @property
    def retweeted(self) -> bool:
        """:class:`bool`: Return True if user retweeted the tweet else False.

        .. versionadded:: 1.2.0
        """
        return self._payload.get("retweeted")


class RelationHide(Relation):
    """Represents a hide relation from a hide & unhide request.

    .. versionadded:: 1.2.0
    """

    __slots__ = ("__original_payload", "_payload")

    def __init__(self, data: Dict[str, Any]):
        self.__original_payload = data
        self._payload = data.get("data")
        super().__init__(4 if self.hidden else None)

    def __repr__(self) -> str:
        return "RelationHide(hidden={0.hidden})".format(self)

    @property
    def hidden(self) -> bool:
        """:class:`bool`: Return True if a tweet is hidden else False

        .. versionadded:: 1.2.0
        """
        return self._payload.get("hidden")


class RelationUpdate(Relation):
    """Represents an update relation that gets return from :meth:`List.update`.

    .. versionadded:: 1.5.0
    """

    __slots__ = ("__original_payload", "_payload")

    def __init__(self, data: Dict[str, Any]):
        self.__original_payload = data
        self._payload = data.get("data")
        super().__init__(5 if self.updated else None)

    def __repr__(self) -> str:
        return "RelationUpdate(updated={0.updated})".format(self)

    @property
    def updated(self) -> bool:
        """:class:`bool`: Return True if the list is updated else False.

        .. versionadded:: 1.5.0
        """
        return self._payload.get("updated")


class RelationPin(Relation):
    """Represents a pin relation that gets return from :meth:`List.pin`.

    .. versionadded:: 1.5.0
    """

    __slots__ = ("__original_payload", "_payload")

    def __init__(self, data: Dict[str, Any]):
        self.__original_payload = data
        self._payload = data.get("data")
        super().__init__(7 if self.pinned else None)

    def __repr__(self) -> str:
        return "RelationPin(pinned={0.pinned})".format(self)

    @property
    def pinned(self) -> bool:
        """:class:`bool`: Return True if the list is pinned else False.

        .. versionadded:: 1.5.0
        """
        return self._payload.get("pinned")


class RelationDelete(Relation):
    """Represents a delete relation that gets return from:

    * :meth:`List.delete`
    * :meth:`Tweet.delete`

    .. versionadded:: 1.5.0
    """

    __slots__ = ("__original_payload", "_payload")

    def __init__(self, data: Dict[str, Any]):
        self.__original_payload = data
        self._payload = data.get("data")
        super().__init__(8 if self.deleted else None)

    def __repr__(self) -> str:
        return "RelationDelete(deleted={0.deleted})".format(self)

    @property
    def deleted(self) -> bool:
        """:class:`bool`: Return True if the list or tweet is deleted else False.

        .. versionadded:: 1.5.0
        """
        return self._payload.get("deleted")



"""
The MIT License (MIT)

Copyright (c) 2021-present UnrealFar & TheGenocides
Copyright (c) 2023-present Sengolda

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""

from __future__ import annotations
from typing import TYPE_CHECKING

from .events import (
    DirectMessageTypingEvent,
    DirectMessageReadEvent,
    TweetFavoriteActionEvent,
    UserRevokeEvent,
    UserFollowActionEvent,
    UserUnfollowActionEvent,
    UserBlockActionEvent,
    UserUnblockActionEvent,
    UserMuteActionEvent,
    UserUnmuteActionEvent,
)
from .message import Message, DirectMessage
from .user import User
from .tweet import Tweet
from .dataclass import TimezoneInfo, Location, SleepTimeSettings, ApplicationInfo
from .utils import convert

if TYPE_CHECKING:
    from .type import Payload
    from .http import HTTPClient

__all__ = ("PayloadParser", "EventParser")


class PayloadParser:
    __slots__ = "http_client"

    def __init__(self, http_client: HTTPClient):
        self.http_client = http_client

    def parse_user_payload(self, payload: Payload):
        copy = payload.copy()
        copy["public_metrics"] = {
            "followers_count": copy.get("followers_count"),
            "following_count": copy.get("friends_count"),
            "tweet_count": copy.get("statuses_count"),
            "listed_count": 0,
        }
        if "created_timestamp" in copy.keys():
            copy["created_at"] = copy.get("created_timestamp")
        if "screen_name" in copy.keys():
            copy["username"] = copy.get("screen_name")

        if "profile_image_url_https" in copy.keys():
            copy["profile_image_url"] = copy.get("profile_image_url_https")
        return copy

    def parse_tweet_payload(self, payload: Payload):
        payload["public_metrics"] = {
            "quote_count": payload.get("quote_count"),
            "reply_count": payload.get("reply_count"),
            "retweet_count": payload.get("retweet_count"),
            "like_count": payload.get("favorite_count"),
        }
        payload["includes"] = {}
        payload["includes"]["mentions"] = [
            user.get("screen_name") for user in payload.get("entities").get("user_mentions")
        ]

        if "timestamp_ms" in payload.keys():
            payload["timestamp"] = payload.get("timestamp_ms")

        if "user" in payload.keys():
            payload["includes"]["users"] = [self.parse_user_payload(payload.get("user"))]

        return payload

    def parse_time_zone_payload(self, payload: Payload):
        payload["name_info"] = payload.get("tzinfo_name")
        payload["timezone"] = TimezoneInfo(**payload.get("time_zone"))
        payload.pop("time_zone")
        payload.pop("tzinfo_name")
        return payload

    def parse_trend_location_payload(self, payload: Payload):
        payload["place_type"] = payload["placeType"]
        payload["country_code"] = payload["countryCode"]
        payload.pop("placeType")
        payload.pop("countryCode")
        payload.pop("parentid")
        payload["location"] = Location(**payload)
        return payload

    def parse_sleep_time_payload(self, payload: Payload):
        payload["sleep_time_setting"] = SleepTimeSettings(**payload["sleep_time"])
        payload.pop("sleep_time")

    def insert_object_author(self, payload: Payload, author: User) -> Payload:
        payload["includes"] = {}
        payload["includes"]["users"] = [author._User__original_payload]
        return payload

    def insert_pagination_object_author(self, payload: Payload) -> list:
        fulldata = []
        for index, data in enumerate(payload["data"]):
            fulldata.append({})
            fulldata[index]["data"] = data
            fulldata[index]["includes"] = {}
            fulldata[index]["includes"]["users"] = [payload.get("includes", {}).get("users", [None])[0]]
        return fulldata

    def parse_message_to_pagination_data(self, data: Payload):
        data["meta"] = {"next_token": data.get("next_cursor"), "previous_token": data.get("previous_cursor")}

        if data.get("events"):
            events = data["events"]
            del data["events"]
            events = data["data"] = events

        else:
            events = data["data"]

        ids = []
        for event_data in events:
            message_create = event_data["message_create"]
            ids.append(message_create["target"]["recipient_id"])
            ids.append(message_create["sender_id"])
        users = self.http_client.fetch_users(ids)

        for event_data in events:
            message_create = event_data["message_create"]
            for user in users:
                if user.id == int(message_create["target"]["recipient_id"]):
                    message_create["target"]["recipient"] = user

                if user.id == int(message_create["sender_id"]):
                    message_create["target"]["sender"] = user

            if data.get("apps"):
                for app in [ApplicationInfo(**data) for data in list(data.get("apps").values())]:
                    if app.id == int(message_create.get("source_app_id", 0)):
                        message_create["target"]["source_application"] = app

                    if app.id == int(message_create.get("source_app_id", 0)):
                        message_create["target"]["source_application"] = app
        return data

    def parse_embed_data(self, payload: Payload) -> Payload:
        payload["status_code"] = payload.get("status") or payload.get("status_code")
        if payload.get("status"):
            del payload["status"]
        return payload

    def parse_metric_data(self, payload: Payload) -> Payload:
        d = {}
        for k, v in payload.items():
            d[k] = convert(v, int)
        return d


class EventParser:
    __slots__ = ("payload_parser", "http_client", "client_id")

    def __init__(self, http_client: HTTPClient):
        self.payload_parser = PayloadParser(http_client)
        self.http_client = http_client
        try:
            self.client_id = int(self.http_client.access_token.partition("-")[0])
        except AttributeError:
            self.client_id = int(self.http_client.fetch_me().id)

    def parse_direct_message_create(self, direct_message_payload: Payload):
        event_payload = {"event": direct_message_payload.get("direct_message_events")[0]}
        users = direct_message_payload.get("users")

        message_create = event_payload["event"].get("message_create")
        recipient_id = message_create.get("target").get("recipient_id")
        sender_id = message_create.get("sender_id")

        recipient = User(
            self.payload_parser.parse_user_payload(users.get(recipient_id)),
            http_client=self.http_client,
        )
        sender = User(
            self.payload_parser.parse_user_payload(users.get(sender_id)),
            http_client=self.http_client,
        )
        application_info = direct_message_payload.get("apps")
        if application_info:
            source_app_id = list(application_info.keys())[0]
            source_app = ApplicationInfo(**application_info.get(source_app_id))

        else:
            source_app = None

        event_payload["event"]["message_create"]["target"]["recipient"] = recipient
        event_payload["event"]["message_create"]["target"]["sender"] = sender
        event_payload["event"]["message_create"]["target"]["source_application"] = source_app

        direct_message = DirectMessage(event_payload, http_client=self.http_client)

        if recipient.id != self.client_id:
            self.http_client.user_cache[recipient.id] = recipient

        if sender.id != self.client_id:
            self.http_client.user_cache[sender.id] = sender

        self.http_client.message_cache[direct_message.id] = direct_message
        self.http_client.dispatch("direct_message", direct_message)

    def parse_direct_message_typing(self, typing_payload: Payload):
        event_payload = typing_payload.get("direct_message_indicate_typing_events")[0]
        users = typing_payload.get("users")

        recipient_id = event_payload.get("target").get("recipient_id")
        sender_id = event_payload.get("sender_id")
        recipient = User(
            self.payload_parser.parse_user_payload(users.get(recipient_id)),
            http_client=self.http_client,
        )
        sender = User(
            self.payload_parser.parse_user_payload(users.get(sender_id)),
            http_client=self.http_client,
        )

        event_payload["target"]["recipient"] = recipient
        event_payload["target"]["sender"] = sender

        if recipient.id != self.client_id:
            self.http_client.user_cache[recipient.id] = recipient

        if sender.id != self.client_id:
            self.http_client.user_cache[sender.id] = sender

        payload = DirectMessageTypingEvent(event_payload, http_client=self.http_client)
        self.http_client.dispatch("typing", payload)

    def parse_direct_message_read(self, read_payload: Payload):
        event_payload = read_payload.get("direct_message_mark_read_events")[0]
        users = read_payload.get("users")

        recipient_id = event_payload.get("target").get("recipient_id")
        sender_id = event_payload.get("sender_id")
        recipient = User(
            self.payload_parser.parse_user_payload(users.get(recipient_id)),
            http_client=self.http_client,
        )
        sender = User(
            self.payload_parser.parse_user_payload(users.get(sender_id)),
            http_client=self.http_client,
        )

        event_payload["target"]["recipient"] = recipient
        event_payload["target"]["sender"] = sender

        if recipient.id != self.client_id:
            self.http_client.user_cache[recipient.id] = recipient

        if sender.id != self.client_id:
            self.http_client.user_cache[sender.id] = sender

        payload = DirectMessageReadEvent(event_payload, http_client=self.http_client)
        self.http_client.dispatch("read", payload)

    def parse_user_revoke(self, action_payload: Payload):
        action = UserRevokeEvent(action_payload)
        self.http_client.dispatch("user_revoke", action)

    def parse_user_action(self, action_payload: Payload, action_type):
        action_payload = action_payload.copy()
        event_payload = action_payload.get(action_type)[0]
        action_type = event_payload.get("type")
        target = User(
            self.payload_parser.parse_user_payload(event_payload.get("target")),
            http_client=self.http_client,
        )
        source = User(
            self.payload_parser.parse_user_payload(event_payload.get("source")),
            http_client=self.http_client,
        )

        event_payload["target"] = target
        event_payload["source"] = source

        if target.id != self.client_id:
            self.http_client.user_cache[target.id] = target

        if source.id != self.client_id:
            self.http_client.user_cache[source.id] = source

        if action_type == "follow":
            action = UserFollowActionEvent(action_payload)
            self.http_client.dispatch("user_follow", action)

        elif action_type == "unfollow":
            action = UserUnfollowActionEvent(action_payload)
            self.http_client.dispatch("user_unfollow", action)

        elif action_type == "block":
            action = UserBlockActionEvent(action_payload)
            self.http_client.dispatch("user_block", action)

        elif action_type == "unblock":
            action = UserUnblockActionEvent(action_payload)
            self.http_client.dispatch("user_unblock", action)

        elif action_type == "mute":
            action = UserMuteActionEvent(action_payload)
            self.http_client.dispatch("user_mute", action)

        elif action_type == "unmute":
            action = UserUnmuteActionEvent(action_payload)
            self.http_client.dispatch("user_unmute", action)

    def parse_tweet_create(self, tweet_payload: Payload):
        tweet_payload = self.payload_parser.parse_tweet_payload(tweet_payload.get("tweet_create_events")[0])
        tweet = Tweet(tweet_payload, http_client=self.http_client)
        self.http_client.tweet_cache[tweet.id] = tweet
        self.http_client.dispatch("tweet_create", tweet)

    def parse_tweet_delete(self, tweet_payload: Payload):
        event_payload = tweet_payload.get("tweet_delete_events")[0]
        tweet_id = event_payload.get("status").get("id")
        tweet = self.http_client.tweet_cache.get(int(tweet_id))
        if not tweet:
            message = Message(None, tweet_id, 1)
            return self.http_client.dispatch("tweet_delete", message)

        try:
            self.http_client.tweet_cache.pop(int(tweet_id))
        except KeyError:
            pass
        finally:
            tweet.deleted_timestamp = int(event_payload.get("timestamp_ms"))
            self.http_client.dispatch("tweet_delete", tweet)

    def parse_favorite_tweet(self, favorite_payload: Payload):
        action_payload = favorite_payload.copy()
        event_payload = favorite_payload.get("favorite_events")[0]
        tweet = Tweet(self.payload_parser.parse_tweet_payload(event_payload.get("favorited_status")))
        user = User(self.payload_parser.parse_user_payload(event_payload.get("user")))
        event_payload["tweet"] = tweet
        event_payload["liker"] = user

        if user.id != self.client_id:
            self.http_client.user_cache[user.id] = user

        action = TweetFavoriteActionEvent(action_payload)
        self.http_client.dispatch("tweet_favorite", action)



"""
The MIT License (MIT)

Copyright (c) 2021-present UnrealFar & TheGenocides
Copyright (c) 2023-present Sengolda

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""

from __future__ import annotations

import json
import requests
import logging
import time

from typing import TYPE_CHECKING, Any, List, Type, Optional
from .dataclass import StreamRule
from .errors import ConnectionException, pytwotException
from .constants import (
    MEDIA_FIELD,
    PLACE_FIELD,
    POLL_FIELD,
    TWEET_EXPANSION,
    TWEET_FIELD,
    USER_FIELD,
)
from .tweet import Tweet


if TYPE_CHECKING:
    from .http import HTTPClient

_log = logging.getLogger(__name__)


class StreamConnection:
    """Represent the twitter api stream connection. This will handle the stream connection.


    .. versionadded:: 1.3.5
    """

    __slots__ = (
        "url",
        "backfill_minutes",
        "reconnect_attempts",
        "http_client",
        "session",
        "errors",
        "running",
    )

    def __init__(
        self,
        url: str,
        backfill_minutes: int = 0,
        reconnect_attempts: int = 0,
        http_client: Optional[HTTPClient] = None,
    ):
        self.url = url
        self.backfill_minutes: int = backfill_minutes
        self.reconnect_attempts: int = reconnect_attempts
        self.http_client: Optional[HTTPClient] = http_client
        self.session: Optional[Any] = None
        self.errors = 0
        self.running = False

    @property
    def closed(self) -> Optional[bool]:
        """Optional[:class:`bool`]: Returns True if the connection is closed, else False.

        .. versionadded:: 1.3.5
        """
        return not self.running

    def is_close(self) -> Optional[bool]:
        """An alias to :class:`StreamConnection.closed`.

        Returns
        ---------
        Optional[:class:`bool`]:
            This method returns a :class:`bool` object.


        .. versionadded:: 1.3.5
        """
        return self.closed

    def close(self) -> None:
        """Close the stream connection.

        Returns
        ---------
        :class:`NoneType`:
            This method returns None.


        .. versionadded:: 1.3.5
        """
        if not self.running:
            raise pytwotException("Attempt to close a stream that's already closed!")

        self.running = False

    def connect(self) -> Optional[Any]:
        """Connect to the current stream connection.


        .. versionadded:: 1.3.5
        """
        self.running = True
        http = self.http_client
        while self.running:
            try:
                response = requests.get(
                    self.url,
                    headers={"Authorization": f"Bearer {http.bearer_token}"},
                    params={
                        "backfill_minutes": int(self.backfill_minutes),
                        "expansions": TWEET_EXPANSION,
                        "media.fields": MEDIA_FIELD,
                        "place.fields": PLACE_FIELD,
                        "poll.fields": POLL_FIELD,
                        "tweet.fields": TWEET_FIELD,
                        "user.fields": USER_FIELD,
                    },
                    stream=True,
                )
                _log.info("Client connected to stream!")
                http.dispatch("stream_connect", self)
                self.session = response

                for response_line in response.iter_lines():
                    if response_line:
                        json_data = json.loads(response_line.decode("UTF-8"))
                        if "errors" in json_data.keys():
                            raise ConnectionException(self.session, None)
                        tweet = Tweet(json_data, http_client=http)
                        http.tweet_cache[tweet.id] = tweet
                        http.dispatch("stream", tweet, self)

            except Exception as e:
                if isinstance(e, AttributeError):
                    break

                elif isinstance(e, requests.exceptions.RequestException):
                    self.errors += 1
                    if self.errors > self.reconnect_attempts:
                        _log.error("Too many errors caught during streaming, closing stream!")
                        self.close()
                        http.dispatch("stream_disconnect", self)
                        break

                    _log.warning(f"An error caught during streaming session: {e}")
                    _log.info(f"Reconnecting to stream after sleeping for 5.0 seconds")
                    time.sleep(5.0)

                else:
                    raise e

        _log.info("Streaming connection has been closed!")


class Stream:
    """Represent a stream object that stream over twitter for tweets.

    Parameters
    ------------
    backfill_minutes: :class:`int`
        This feature will deliver duplicate Tweets, meaning that if you were disconnected for 90 seconds, and you requested two minutes of backfill, you will receive 30 seconds worth of duplicate Tweets. Due to this, you should make sure your system is tolerant of duplicate data. This feature is currently only available to the Academic Research product track.
    reconnect_attempts: :class:`int`
        Decide how many attempts for a reconnect to perform, if the client reconnected more then this argument, it would break the loop.


    .. versionadded:: 1.3.5
    """

    __slots__ = (
        "backfill_minutes",
        "raw_rules",
        "http_client",
        "reconnect_attempts",
        "sample",
        "connection",
    )

    def __init__(self, backfill_minutes: int = 0, reconnect_attempts: int = 15):
        self.backfill_minutes = backfill_minutes
        self.raw_rules: Optional[list] = []
        self.http_client: Optional[HTTPClient] = None
        self.reconnect_attempts = reconnect_attempts
        self.sample = False
        self.connection: StreamConnection = StreamConnection(
            "https://api.twitter.com/2/tweets/search/stream",
            self.backfill_minutes,
            reconnect_attempts,
            self.http_client,
        )

    @classmethod
    def sample_stream(cls: Type[Stream], backfill_minutes: int = 0, reconnect_attempts: int = 15) -> Stream:
        """A class method that change the stream connection to a sample one, this would mean you dont have to set any stream rules. This would not recommended because it can make the progress of tweet cap much faster, if its out of limit you would not be able to stream.

        Parameters
        ------------
        backfill_minutes: :class:`int`
            This feature will deliver duplicate Tweets, meaning that if you were disconnected for 90 seconds, and you requested two minutes of backfill, you will receive 30 seconds worth of duplicate Tweets. Due to this, you should make sure your system is tolerant of duplicate data. This feature is currently only available to the Academic Research product track.
        reconnect_attempts: :class:`int`
            Decide how many attempts for a reconnect to perform, if the client reconnected more then this argument, it would break the loop.

        Returns
        ---------
        :class:`Stream`
            This classmethod returns your :class:`Stream` instance.


        .. versionadded:: 1.3.5
        """
        self = cls(backfill_minutes, reconnect_attempts)
        self.http_client = None
        self.sample = True
        self.connection = StreamConnection(
            "https://api.twitter.com/2/tweets/sample/stream",
            self.backfill_minutes,
            self.reconnect_attempts,
            self.http_client,
        )
        return self

    @property
    def rules(self) -> Optional[dict]:
        """:class:`dict`: Returns the stream's rules, if its a sample stream it would returns None.

        .. versionadded:: 1.3.5
        """
        if self.sample:
            return None

        return [StreamRule(**data) for data in self.raw_rules]

    def add_rule(self, value: str, tag: Optional[str] = None) -> Optional[Stream]:
        """Add a rule to your stream to match with tweets that the stream return. You can use an operator to do this, check https://developer.twitter.com/en/docs/twitter-api/tweets/search/integrate/build-a-query for more information about the operator. You can add multiple rules depending your access level, 5 for essential access, 25 for elevated access and 100 for academic research access.

        Parameters
        ------------
        value: :class:`str`
            The rule text. If you are using a Standard Project at the Basic access level, you can use the basic set of operators, can submit up to 25 concurrent rules, and can submit rules up to 512 characters long. If you are using an Academic Research Project at the Basic access level, you can use all available operators, can submit up to 1,000 concurrent rules, and can submit rules up to 1,024 characters long.
        tag: Optional[:class:`str`]
            The tag label. This is a free-form text you can use to identify the rules that matched a specific Tweet in the streaming response. Tags can be the same across rules.

        Returns
        ---------
        :class:`Stream`:
            Returns the stream's rules, if its a sample stream it would returns None.


        .. versionadded:: 1.3.5
        """
        if self.sample:
            return None

        data = {"value": value}
        if tag:
            data["tag"] = tag
        self.raw_rules.append(data)
        return self

    def clear(self) -> None:
        """Clear stream before attempting to connect with the stream connection, this would delete all previous rules in your stream.

        Returns
        ---------
        :class:`NoneType`:
            This method returns None.


        .. versionadded:: 1.3.5
        """
        if self.sample:
            return None

        rules = self.http_client.request(
            "GET",
            "2",
            "/tweets/search/stream/rules",
        )
        if not rules:
            return

        if rules.get("data"):
            data = {"delete": {"ids": [str(rule.get("id")) for rule in rules.get("data")]}}
            rules = self.http_client.request("POST", "2", "/tweets/search/stream/rules", json=data)

    def fetch_rules(self) -> Optional[List[StreamRule]]:
        """Fetches the stream's rules.

        Returns
        ---------
        Optional[List[:class:`StreamRule`]]
            This method returns a :class:`list` of :class:`StreamRule` objects.


        .. versionadded:: 1.3.5
        """
        if self.sample:
            return None

        res = self.http_client.request(
            "GET",
            "2",
            "/tweets/search/stream/rules",
        )

        try:
            return [StreamRule(**data) for data in res["data"]]
        except TypeError:
            return res

    def set_rules(self, dry_run: bool) -> None:
        """Create and set rules to your stream.

        Parameters
        ------------
        dry_run: :class:`bool`
            Indicates if you want to debug your rule's operator syntax.


        .. versionadded:: 1.3.5
        """
        if self.sample:
            return None

        if dry_run:
            try:
                self.http_client.request(
                    "POST",
                    "2",
                    "/tweets/search/stream/rules",
                    params={"dry_run": dry_run},
                    json={"add": self.raw_rules},
                )
            except pytwotException as e:
                raise e
            else:
                self.http_client.request(
                    "POST",
                    "2",
                    "/tweets/search/stream/rules",
                    json={"add": self.raw_rules},
                )
                return

        self.http_client.request("POST", "2", "/tweets/search/stream/rules", json={"add": self.raw_rules})

    def connect(self, *, dry_run: bool = False) -> None:
        """Connect with the stream connection.

        Parameters
        ------------
        dry_run: :class:`bool`
            Indicates if you want to debug your rule's operator syntax. Default to None.


        .. versionadded:: 1.3.5
        """
        self.clear()
        self.set_rules(dry_run)
        self.connection.connect()



"""
The MIT License (MIT)

Copyright (c) 2021-present UnrealFar & TheGenocides
Copyright (c) 2023-present Sengolda

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""

from __future__ import annotations

import threading
import random
import string
from concurrent.futures import ThreadPoolExecutor, wait
from typing import Optional, Callable, Any
from ..constants import ALL_COMPLETED

__all__ = ("Executor", "ThreadManager")


class Executor(ThreadPoolExecutor):
    def __init__(self, *args, **kwargs):
        self._session_id = kwargs.pop("session_id")
        self._thread_name = kwargs.pop("thread_name")
        self._futures = []
        self._thread_name += f":session_id={self.session_id}:task_number="
        super().__init__(thread_name_prefix=self._thread_name, *args, **kwargs)

    @property
    def futures(self):
        return self._futures

    @property
    def threads(self):
        return self._threads

    @property
    def session_id(self):
        return self._session_id

    def submit(self, fn: Callable, *args: Any, **kwargs: Any):
        future = super().submit(fn, *args, **kwargs)
        self.futures.append(future)
        return future

    def clear_futures(self):
        self.futures.clear()

    def wait_for_futures(
        self,
        *,
        timeout: Optional[int] = None,
        return_when=ALL_COMPLETED,
        purge: bool = True,
    ):
        if not self.futures:
            return None
        result = wait(self.futures, timeout, return_when)
        if purge:
            self.clear_futures()
        return result


class ThreadManager:
    @property
    def active_threads(self) -> list:
        return threading.enumerate()

    def create_new_executor(self, *, max_workers: int = 100, thread_name: str = "", session_id: str = None) -> Executor:
        return Executor(
            max_workers,
            thread_name=thread_name,
            session_id=session_id or self.generate_thread_session(),
        )

    def get_threads(self, session_id):
        threads = []
        for t in self.active_threads:
            if session_id in t.name:
                threads.append(t)
        return threads

    def generate_thread_session(self):
        return "".join((random.sample(string.ascii_lowercase, 10)))



"""
The MIT License (MIT)

Copyright (c) 2021-present UnrealFar & TheGenocides
Copyright (c) 2023-present Sengolda

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""

from __future__ import annotations

from dataclasses import dataclass
from typing import Optional


@dataclass
class Location:
    """Represents a Location that Twitter has trending topic information for.

    .. versionadded:: 1.5.0
    """

    country: str
    country_code: str
    name: str
    parent_id: int
    place_type: PlaceType
    url: str
    woeid: int


@dataclass
class Trend:
    """Represents a twitter's trending topics display as Trend.

    .. versionadded:: 1.5.0
    """

    name: str
    url: str
    promoted_content: Optional[str]
    query: str
    tweet_volume: Optional[str]


@dataclass
class TimezoneInfo:
    """Represents a TimezoneInfo for the user's setting.

    .. versionadded:: 1.5.0
    """

    name: str
    name_info: str
    utc_offset: int


@dataclass
class PlaceType:
    """Represents a place type consist of the code and name for :class:`Location`.

    .. versionadded:: 1.5.0
    """

    code: int
    name: str



"""
The MIT License (MIT)

Copyright (c) 2021-present UnrealFar & TheGenocides
Copyright (c) 2023-present Sengolda

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""

from __future__ import annotations
from typing import Any, List, Tuple, Optional, TYPE_CHECKING
from .errors import NoPageAvailable


if TYPE_CHECKING:
    from .http import HTTPClient
    from .type import Payload


class Pagination:
    """Represents the base class of all pagination objects.


    .. versionadded:: 1.5.0
    """

    __slots__ = (
        "__original_payload",
        "_payload",
        "_meta",
        "_next_token",
        "_previous_token",
        "_count",
        "_paginate_over",
        "_current_page_number",
        "_params",
        "item_type",
        "endpoint_request",
        "http_client",
        "pages_cache",
    )

    def __init__(
        self,
        data: Payload,
        *,
        item_type: Any,
        endpoint_request: str,
        http_client: HTTPClient,
        **kwargs: Any,
    ):
        self.__original_payload = data
        self._payload = self.__original_payload.get("data")
        self._meta = self.__original_payload.get("meta")
        self._next_token = self._meta.get("next_token")
        self._previous_token = self._meta.get("previous_token")
        self._count = 0
        self._paginate_over = 0
        self._current_page_number = 1
        self._params = kwargs.get("params", None)
        self.item_type = item_type
        self.endpoint_request = endpoint_request
        self.http_client = http_client
        self.pages_cache = {1: {obj.id: obj for obj in self.content}}

    @property
    def original_payload(self):
        return self.__original_payload

    @original_payload.setter
    def original_payload(self, other: dict):
        self.__original_payload = other
        return self.original_payload

    @property
    def payload(self):
        return self._payload

    @payload.setter
    def payload(self, other: dict):
        self._payload = other
        return self._payload

    @property
    def content(self) -> list:
        """:class:`list`: Returns a list of objects from the current page's content.

        .. versionadded:: 1.5.0
        """
        return [self.item_type(data, http_client=self.http_client) for data in self.payload]

    @property
    def paginate_over(self) -> int:
        """:class:`int`: Returns how many times you change page over the pagination.

        .. versionadded:: 1.5.0
        """
        return self._paginate_over

    @property
    def current_page_number(self) -> int:
        """:class:`int`: Returns the current page number.

        .. versionadded:: 1.5.0
        """
        return self._current_page_number

    @property
    def pages(self) -> List[Tuple[int, list]]:
        """List[Tuple[:class:`int`, :class:`list`]]: Returns the zipped pages with the page number and content from a cache. If you never been into the page you want, it might not be returns in this property. example to use:

        .. code-block:: py

            for page_number, page_content in pagination.pages:
                ... #do something


        .. versionadded:: 1.5.0
        """
        fulldata = []
        for page_number in self.pages_cache.keys():
            fulldata.append(list(self.pages_cache.get(page_number).values()))
        return zip(range(1, len(self.pages_cache) + 1), fulldata)

    def get_page_content(self, page_number: int) -> Optional[list]:
        """Gets the page `content` from the pagination pages cache. If you never been into the page you want, it might not be returns.

        .. note::
            Note that, if the page_number is 0 it automatically would returns None. Specify number 1 or above.

        Returns
        ---------
        Optional[:class:`list`]
            This method returns a list of objects.


        .. versionadded:: 1.5.0
        """
        content = self.pages_cache.get(page_number)
        if not content:
            return None
        return list(content.values())

    def next_page(self):
        raise NotImplementedError

    def previous_page(self):
        raise NotImplementedError


class UserPagination(Pagination):
    """Represents a pagination that handles users object. This inherits :class:`Pagination`. These following methods return this object:

    * :meth:`User.fetch_following`
    * :meth:`User.fetch_followers`
    * :meth:`User.fetch_muters`
    * :meth:`User.fetch_blockers`
    * :meth:`Tweet.fetch_likers`
    * :meth:`Tweet.fetch_retweeters`
    * :meth:`List.fetch_members`


    .. versionadded:: 1.5.0
    """

    def __init__(self, data, **kwargs):
        from .user import User  # Avoid circular import error.

        super().__init__(data, item_type=User, **kwargs)

    def next_page(self):
        """Change `content` property to the next page's contents..

        Raises
        --------
        :class:`NoPageAvailable`
            Raises when you reached the end of the pagination.


        .. versionadded:: 1.5.0
        """
        if not self._next_token:
            raise NoPageAvailable()
        self._params["pagination_token"] = self._next_token

        res = self.http_client.request(
            "GET",
            "2",
            self.endpoint_request,
            auth=True,
            params=self._params,
        )
        if not res:
            raise NoPageAvailable()

        previous_content = self.content
        self._current_page_number += 1
        self.original_payload = res
        self.payload = self.original_payload.get("data")
        self._meta = self.original_payload.get("meta")
        self._next_token = self._meta.get("next_token")
        self._previous_token = self._meta.get("previous_token")
        self._count = 0

        if not previous_content[0] == self.content[0]:
            self.pages_cache[len(self.pages_cache) + 1] = {user.id: user for user in self.content}

    def previous_page(self):
        """Change `content` property to the previous page's contents..

        Raises
        --------
        :class:`NoPageAvailable`
            Raises when you reached the end of the pagination.


        .. versionadded:: 1.5.0
        """
        if not self._previous_token:
            raise NoPageAvailable()
        self._params["pagination_token"] = self._previous_token

        res = self.http_client.request(
            "GET",
            "2",
            self.endpoint_request,
            auth=True,
            params=self._params,
        )
        if not res:
            raise NoPageAvailable()

        previous_content = self.content
        self._current_page_number -= 1
        self.original_payload = res
        self.payload = self.original_payload.get("data")
        self._meta = self.original_payload.get("meta")
        self._next_token = self._meta.get("next_token")
        self._previous_token = self._meta.get("previous_token")
        self._count = 0

        if not previous_content[0] == self.content[0]:
            self.pages_cache[len(self.pages_cache) + 1] = {user.id: user for user in self.content}


class TweetPagination(Pagination):
    """Represents a pagination that handles tweets object. This inherits :class:`Pagination`. These following methods return this object:

    * meth:`User.fetch_timelines`
    * meth:`User.fetch_liked_tweets`
    * meth:`List.fetch_tweets`


    .. versionadded:: 1.5.0
    """

    def __init__(self, data, **kwargs):
        from .tweet import Tweet  # Avoid circular import error.

        super().__init__(data, item_type=Tweet, **kwargs)

    @property
    def content(self) -> list:
        """:class:`list`: Returns a list of objects from the current page's content.

        .. versionadded:: 1.5.0
        """

        return [
            self.item_type(data, http_client=self.http_client)
            for data in self.http_client.payload_parser.insert_pagination_object_author(self.original_payload)
        ]

    def next_page(self):
        """Change `content` property to the next page's contents..

        Raises
        --------
        :class:`NoPageAvailable`
            Raises when you reached the end of the pagination.


        .. versionadded:: 1.5.0
        """
        if not self._next_token:
            raise NoPageAvailable()
        self._params["pagination_token"] = self._next_token

        res = self.http_client.request(
            "GET",
            "2",
            self.endpoint_request,
            auth=True,
            params=self._params,
        )
        if not res:
            raise NoPageAvailable()

        previous_content = self.content
        self._current_page_number += 1
        self.original_payload = res
        self.payload = self.content
        self._meta = self.original_payload.get("meta")
        self._next_token = self._meta.get("next_token")
        self._previous_token = self._meta.get("previous_token")
        self._count = 0

        if not previous_content[0] == self.content[0]:
            self.pages_cache[len(self.pages_cache) + 1] = {tweet.id: tweet for tweet in self.content}

    def previous_page(self):
        """Change `content` property to the previous page's contents..

        Raises
        --------
        :class:`NoPageAvailable`
            Raises when you reached the end of the pagination.


        .. versionadded:: 1.5.0
        """
        if not self._previous_token:
            raise NoPageAvailable()
        self._params["pagination_token"] = self._previous_token

        res = self.http_client.request(
            "GET",
            "2",
            self.endpoint_request,
            auth=True,
            params=self._params,
        )
        if not res:
            raise NoPageAvailable()

        previous_content = self.content
        self._current_page_number -= 1
        self.original_payload = res
        self.payload = self.content
        self._meta = self.original_payload.get("meta")
        self._next_token = self._meta.get("next_token")
        self._previous_token = self._meta.get("previous_token")
        self._count = 0

        if not previous_content[0] == self.content[0]:
            self.pages_cache[len(self.pages_cache) + 1] = {tweet.id: tweet for tweet in self.content}


class ListPagination(Pagination):
    """Represents a pagination that handles list objects. This inherits :class:`Pagination`. These following methods return this object:

    * :meth:`User.fetch_lists`
    * :meth:`User.fetch_list_memberships`


    .. versionadded:: 1.5.0
    """

    def __init__(self, data, **kwargs):
        from .list import List as TwitterList  # Avoid circular import error

        super().__init__(data, item_type=TwitterList, **kwargs)

    @property
    def content(self) -> list:
        """:class:`list`: Returns a list of objects from the current page's content.

        .. versionadded:: 1.5.0
        """

        return [
            self.item_type(data, http_client=self.http_client)
            for data in self.http_client.payload_parser.insert_pagination_object_author(self.original_payload)
        ]

    def next_page(self):
        """Change `content` property to the next page's contents..

        Raises
        --------
        :class:`NoPageAvailable`
            Raises when you reached the end of the pagination.


        .. versionadded:: 1.5.0
        """
        if not self._next_token:
            raise NoPageAvailable()
        self._params["pagination_token"] = self._next_token

        res = self.http_client.request(
            "GET",
            "2",
            self.endpoint_request,
            auth=True,
            params=self._params,
        )
        if not res:
            raise NoPageAvailable()

        previous_content = self.content
        self._current_page_number += 1
        self.original_payload = res
        self.payload = self.content
        self._meta = self.original_payload.get("meta")
        self._next_token = self._meta.get("next_token")
        self._previous_token = self._meta.get("previous_token")
        self._count = 0

        if not previous_content[0] == self.content[0]:
            self.pages_cache[len(self.pages_cache) + 1] = {
                _TwitterList.id: _TwitterList for _TwitterList in self.content
            }

    def previous_page(self):
        """Change `content` property to the previous page's contents..

        Raises
        --------
        :class:`NoPageAvailable`
            Raises when you reached the end of the pagination.


        .. versionadded:: 1.5.0
        """
        if not self._previous_token:
            raise NoPageAvailable()
        self._params["pagination_token"] = self._previous_token

        res = self.http_client.request(
            "GET",
            "2",
            self.endpoint_request,
            auth=True,
            params=self._params,
        )
        if not res:
            raise NoPageAvailable()

        previous_content = self.content
        self._current_page_number -= 1
        self.original_payload = res
        self.payload = self.content
        self._meta = self.original_payload.get("meta")
        self._next_token = self._meta.get("next_token")
        self._previous_token = self._meta.get("previous_token")
        self._count = 0

        if not previous_content[0] == self.content[0]:
            self.pages_cache[len(self.pages_cache) + 1] = {
                _TwitterList.id: _TwitterList for _TwitterList in self.content
            }


class MessagePagination(Pagination):
    """ "Represents a pagination for message objects.
    These methods returns this pagination object:

    * :meth:`Client.fetch_message_history`


    .. versionadded:: 1.5.0
    """

    def __init__(self, data, **kwargs):
        from .message import DirectMessage  # Avoid circular import error.

        data = kwargs.get("http_client").payload_parser.parse_message_to_pagination_data(data)
        super().__init__(data, item_type=DirectMessage, **kwargs)

    def next_page(self):
        """Change `content` property to the next page's contents..

        Raises
        --------
        :class:`NoPageAvailable`
            Raises when you reached the end of the pagination.


        .. versionadded:: 1.5.0
        """
        if not self._next_token:
            raise NoPageAvailable()
        self._params["cursor"] = self._next_token

        res = self.http_client.request(
            "GET",
            "1.1",
            self.endpoint_request,
            auth=True,
            params=self._params,
        )
        if not res:
            raise NoPageAvailable()

        previous_content = self.content
        self._current_page_number += 1
        self.original_payload = self.http_client.payload_parser.parse_message_to_pagination_data(res)
        self.payload = self.content
        self._meta = self.original_payload.get("meta")
        self._next_token = self._meta.get("next_token")
        self._previous_token = self._meta.get("previous_token")
        self._count = 0

        if not previous_content[0] == self.content[0]:
            self.pages_cache[len(self.pages_cache) + 1] = {message.id: message for message in self.content}

    def previous_page(self):
        """Change `content` property to the previous page's contents..

        Raises
        --------
        :class:`NoPageAvailable`
            Raises when you reached the end of the pagination.


        .. versionadded:: 1.5.0
        """
        if not self._previous_token:
            raise NoPageAvailable()
        self._params["cursor"] = self._previous_token

        res = self.http_client.request(
            "GET",
            "1.1",
            self.endpoint_request,
            auth=True,
            params=self._params,
        )
        if not res:
            raise NoPageAvailable()

        previous_content = self.content
        self._current_page_number -= 1
        self.original_payload = self.http_client.payload_parser.parse_message_to_pagination_data(res)
        self.payload = self.content
        self._meta = self.original_payload.get("meta")
        self._next_token = self._meta.get("next_token")
        self._previous_token = self._meta.get("previous_token")
        self._count = 0

        if not previous_content[0] == self.content[0]:
            self.pages_cache[len(self.pages_cache) + 1] = {message.id: message for message in self.content}



from .attachments import *
from .locations import *
from .message import *
from .metrics import *
from .settings import *
from .space import *
from .stream import *
from .app import *
from .compliance import *
from .embed import *



"""
The MIT License (MIT)

Copyright (c) 2021-present UnrealFar & TheGenocides
Copyright (c) 2023-present Sengolda

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""

from __future__ import annotations

import io
import logging
import sys
import time
import requests
from json import JSONDecodeError
from typing import Any, List, NoReturn, Optional, Union, TYPE_CHECKING

from .attachments import CTA, CustomProfile, File, Geo, Poll, QuickReply
from .auth import OauthSession
from .enums import ReplySetting, SpaceState, Granularity, JobType, JobStatus
from .errors import (
    BadRequests,
    Conflict,
    Forbidden,
    NotFound,
    TooManyRequests,
    pytwotException,
    Unauthorized,
    FieldsTooLarge,
    UnauthorizedForResource,
    ResourceNotFound,
    DisallowedResource,
)
from .constants import (
    TWEET_EXPANSION,
    SPACE_EXPANSION,
    LIST_EXPANSION,
    PINNED_TWEET_EXPANSION,
    MEDIA_FIELD,
    PLACE_FIELD,
    POLL_FIELD,
    SPACE_FIELD,
    COMPLETE_SPACE_FIELD,
    TWEET_FIELD,
    COMPLETE_TWEET_FIELD,
    TWEET_FIELD_WITH_ORGANIC_METRICS,
    TWEET_FIELD_WITH_PROMOTED_METRICS,
    USER_FIELD,
    TOPIC_FIELD,
    LIST_FIELD,
)
from .message import DirectMessage, WelcomeMessage, WelcomeMessageRule
from .parser import EventParser
from .space import Space
from .tweet import Tweet
from .user import User, ClientAccount
from .threads import ThreadManager
from .relations import RelationUpdate
from .list import List as TwitterList
from .compliance import Job

if TYPE_CHECKING:
    from .type import ID, Payload, ResponsePayload
    from .stream import Stream

_log = logging.getLogger(__name__)


class HTTPClient:
    def __init__(
        self,
        bearer_token: str,
        *,
        consumer_key: Optional[str],
        consumer_secret: Optional[str],
        access_token: Optional[str],
        access_token_secret: Optional[str],
        stream: Optional[Stream] = None,
        callback_url: Optional[str] = None,
        client_id: Optional[str] = None,
        client_secret: Optional[str] = None,
        use_bearer_only: bool = False,
        sleep_after_ratelimit: bool = False,
    ):
        self.credentials = {
            "bearer_token": bearer_token,
            "consumer_key": consumer_key,
            "consumer_secret": consumer_secret,
            "access_token": access_token,
            "access_token_secret": access_token_secret,
        }
        if not bearer_token:
            _log.error("bearer token is missing!")
        if not consumer_key:
            _log.warning("Consumer key is missing this is recommended to have!")
        if not access_token:
            _log.warning("Access token is missing this is recommended to have")
        if not access_token_secret:
            _log.warning(
                "Access token secret is missing this is required if you have passed in the access_token param."
            )

        for k, v in self.credentials.items():
            if not isinstance(v, (str, type(None))):
                raise Unauthorized(None, f"Wrong authorization passed for credential: {k}.")

        self.__session = requests.Session()
        self.base_url = "https://api.twitter.com/"
        self.upload_url = "https://upload.twitter.com/"
        self.bearer_token = bearer_token
        self.consumer_key = consumer_key
        self.consumer_secret = consumer_secret
        self.access_token = access_token
        self.access_token_secret = access_token_secret
        self.stream = stream
        self.callback_url = callback_url
        self.client_id = client_id
        self.client_secret = client_secret
        self._auth = OauthSession(
            self.consumer_key,
            self.consumer_secret,
            access_token=self.access_token,
            access_token_secret=self.access_token_secret,
            http_client=self,
            callback_url=self.callback_url,
            client_id=self.client_id,
            client_secret=self.client_secret,
        )
        self.use_bearer_only = use_bearer_only
        self.event_parser = EventParser(self)
        self.payload_parser = self.event_parser.payload_parser
        self.thread_manager = ThreadManager()
        self.sleep_after_ratelimit = sleep_after_ratelimit
        self.current_header = None
        self.message_cache = {}
        self.tweet_cache = {}
        self.user_cache = {}
        self.events = {}
        if self.stream:
            self.stream.http_client = self
            self.stream.connection.http_client = self

    @property
    def access_levels(self) -> Optional[list]:
        return self.current_header.get("x-access-levels").split("-") if self.current_header else None

    @property
    def oauth_session(self) -> OauthSession:
        return self._auth

    def dispatch(self, event_name: str, *args: Any, **kwargs: Any) -> Any:
        event = self.events.get(event_name)
        if not event:
            return None

        _log.debug(f"Dispatching Event: on_{event_name}" f"Positional-Arguments: {args}" f"Keyword-Arguments: {kwargs}")
        return event(*args, **kwargs)

    def request(
        self,
        method: str,
        version: str,
        path: str,
        *,
        headers: Payload = {},
        params: Payload = {},
        json: Payload = {},
        data: Payload = {},
        files: Payload = {},
        auth: bool = False,
        basic_auth: bool = False,
        thread_name: Optional[str] = None,
        thread_session: bool = False,
        use_base_url: bool = True,
        is_json: bool = True,
    ) -> ResponsePayload:
        if use_base_url:
            url = self.base_url + version + path
        else:
            url = self.upload_url + version + path

        user_agent = "Py-Tweet (https://github.com/sengolda/pytwot/) Python/{0[0]}.{0[1]}.{0[2]} requests/{1}"
        if "Authorization" not in headers.keys():
            headers["Authorization"] = f"Bearer {self.bearer_token}"

        headers["User-Agent"] = user_agent.format(sys.version_info, requests.__version__)

        if not self.use_bearer_only:
            if auth:
                auth = self.oauth_session.oauth1
                for k, v in self.credentials.items():
                    if v is None:
                        raise pytwotException(f"{k} is a required credential for authorization.")

            if basic_auth:
                headers["Authorization"] = f"Basic {self.oauth_session.basic_auth}"

        else:
            auth = None
            basic_auth = None

        if data:
            json = None

        if json:
            data = None

        method = method.upper()
        if thread_session:
            executor = self.thread_manager.create_new_executor(thread_name=thread_name, session_id=thread_session)
            future = executor.submit(
                self.request,
                method,
                version,
                path,
                headers=headers,
                params=params,
                data=data,
                json=json,
                files=files,
                auth=auth,
            )
            return future

        else:
            response = self.__session.request(
                method,
                url,
                headers=headers,
                params=params,
                data=data,
                json=json,
                files=files,
                auth=auth,
            )
            code = response.status_code
            self.current_header = response.headers
            res = None
            _log.debug(
                f"{method} {url} has returned: "
                f"{response.status_code} {response.reason}\n"
                f"Headers: {response.headers}\n"
                f"Content: {response.content}\n"
                f"Json-payload: {json}\n"
                f"Parameters: {params}\n"
            )

            if code in (201, 202, 204):
                if is_json:
                    try:
                        res = response.json()
                    except JSONDecodeError:
                        return response.text
                else:
                    return response.text

                return res

            elif code == 400:
                raise BadRequests(response)

            elif code == 401:
                raise Unauthorized(response)

            elif code == 403:
                raise Forbidden(response)

            elif code == 404:
                raise NotFound(response)

            elif code == 409:
                raise Conflict(response)

            elif code in (420, 429):  # 420 status code is an unofficial extension by Twitter.
                if self.sleep_after_ratelimit:
                    remaining = int(response.headers["x-rate-limit-reset"])
                    sleep_for = (remaining - int(time.time())) + 1
                    _log.warn(f"Client is ratelimited. Sleeping for {sleep_for}")
                    print(f"Client is ratelimited. Sleeping for {sleep_for}")
                    time.sleep(sleep_for)
                    return self.request(
                        method,
                        version,
                        path,
                        headers=headers,
                        params=params,
                        data=data,
                        json=json,
                        files=files,
                        auth=auth,
                    )

                else:
                    raise TooManyRequests(response)

            elif code == 431:
                raise FieldsTooLarge(response)

            if is_json:
                try:
                    res = response.json()
                except JSONDecodeError:
                    res = response.text
            else:
                return response.text

            if isinstance(res, dict):
                if res.get("errors"):
                    error = res["errors"][0]
                    if error["type"] == "https://api.twitter.com/2/problems/not-authorized-for-resource":
                        if not error["parameter"] == "pinned_tweet_id":
                            raise UnauthorizedForResource(error["detail"])
                    elif (
                        error["type"] == "https://api.twitter.com/2/problems/resource-not-found"
                        and res.get("data", None) is None
                    ):
                        raise ResourceNotFound(error["detail"])

                    elif error["type"] == "https://api.twitter.com/2/problems/disallowed-resource":
                        raise DisallowedResource(error["detail"])

                if "meta" in res.keys():
                    try:
                        if res["meta"]["result_count"] == 0:
                            return []
                    except KeyError:
                        pass

            return res

    def upload(self, file: File, command: str):
        assert command.upper() in ("INIT", "APPEND", "FINALIZE", "STATUS")
        thread_session = self.thread_manager.generate_thread_session()

        def check_status(processing_info, media_id):
            if not processing_info:
                return

            state = processing_info["state"]
            seconds = processing_info.get("check_after_secs")
            if seconds is None:
                return None

            if state == "succeeded":
                return

            if state == "failed":
                raise pytwotException(f"Failed to finalize Media!\n{processing_info}")

            time.sleep(seconds)

            res = self.request(
                "GET",
                version="1.1",
                path="/media/upload.json",
                params={"command": "STATUS", "media_id": media_id},
                auth=True,
                use_base_url=False,
            )

            processing_info = res.get("processing_info", None)
            check_status(processing_info, media_id)

        if command.upper() == "INIT":
            data = {
                "command": "INIT",
                "media_type": file.mimetype,
                "total_bytes": file.total_bytes,
                "media_category": file.media_category,
                "shared": file.dm_only,
            }

            res = self.request(
                "POST",
                "1.1",
                "/media/upload.json",
                data=data,
                auth=True,
                use_base_url=False,
            )

            file._File__media_id = res["media_id"]
            return res["media_id"]

        elif command.upper() == "APPEND":
            segment_id = 0
            bytes_sent = 0
            path = file.path
            if isinstance(path, io.IOBase):
                open_file = path
            else:
                open_file = open(path, "rb")

            if not file.media_id:
                raise ValueError("'media_id' is None! Please specified it.")

            while bytes_sent < file.total_bytes:
                self.request(
                    "POST",
                    version="1.1",
                    path="/media/upload.json",
                    data={
                        "command": "APPEND",
                        "media_id": file.media_id,
                        "segment_index": segment_id,
                    },
                    files={"media": open_file.read(4 * 1024 * 1024)},
                    auth=True,
                    use_base_url=False,
                )

                bytes_sent = open_file.tell()
                segment_id += 1

        elif command.upper() == "FINALIZE":
            executor = self.thread_manager.create_new_executor(thread_name="subfiles-upload-request")
            res = self.request(
                "POST",
                version="1.1",
                path="/media/upload.json",
                data={"command": "FINALIZE", "media_id": file.media_id},
                auth=True,
                use_base_url=False,
            )

            check_status(res.get("processing_info", None), file.media_id)
            if file.alt_text:
                alt_text_future = self.request(
                    "POST",
                    "1.1",
                    "/media/metadata/create.json",
                    json={
                        "media_id": str(file.media_id),
                        "alt_text": {"text": str(file.alt_text)},
                    },
                    auth=True,
                    use_base_url=False,
                    thread_name="alt-text-file-request",
                    thread_session=thread_session,
                )

            if file.subfile:
                executor.submit(self.quick_upload, file.subfile)

            if file.subfiles:
                for subfile in file.subfiles:
                    executor.submit(self.quick_upload, subfile)

            if file.subfiles or file.subfile:
                subtitles = []
                executor.wait_for_futures()
                if file.subfiles:
                    for subfile in file.subfiles:
                        subtitles.append(
                            {
                                "media_id": str(subfile.media_id),
                                "display_name": subfile.language,
                                "language_code": subfile.language_code,
                            }
                        )

                if file.subfile:
                    subtitles.append(
                        {
                            "media_id": str(file.subfile.media_id),
                            "display_name": file.subfile.language,
                            "language_code": file.subfile.language_code,
                        }
                    )

                subtitle_data = {
                    "media_id": str(file.media_id),
                    "media_category": file.media_category,
                    "subtitle_info": {"subtitles": subtitles},
                }

                self.request(
                    "POST",
                    version="1.1",
                    path="/media/subtitles/create.json",
                    json=subtitle_data,
                    auth=True,
                    use_base_url=False,
                    thread_name=f"subfile-request:FILE-MEDIA-ID={file.media_id}:SUBFILE-MEDIA-ID={file.subfile.media_id}",
                )

            if file.alt_text:
                alt_text_future.result()

    def quick_upload(self, file: File) -> File:
        self.upload(file, "INIT")
        self.upload(file, "APPEND")
        self.upload(file, "FINALIZE")
        return file

    def fetch_me(self):
        data = self.request(
            "GET",
            "2",
            f"/users/me",
            params={
                "expansions": PINNED_TWEET_EXPANSION,
                "user.fields": USER_FIELD,
                "tweet.fields": TWEET_FIELD,
            },
            auth=True,
        )

        return User(data, http_client=self)

    def fetch_user(self, user_id: ID) -> Optional[User]:
        try:
            int(user_id)
        except ValueError:
            raise ValueError("user_id must be an int, or a string of digits!")

        data = self.request(
            "GET",
            "2",
            f"/users/{user_id}",
            params={
                "expansions": PINNED_TWEET_EXPANSION,
                "user.fields": USER_FIELD,
                "tweet.fields": TWEET_FIELD,
            },
            auth=True,
        )

        return User(data, http_client=self)

    def fetch_users(self, ids: List[ID]) -> List[User]:
        str_ids = []
        for id in ids:
            try:
                int(id)
            except ValueError as e:
                raise e
            else:
                str_ids.append(str(id))

        res = self.request(
            "GET",
            "2",
            f"/users?ids={','.join(str_ids)}",
            params={
                "expansions": PINNED_TWEET_EXPANSION,
                "user.fields": USER_FIELD,
                "tweet.fields": TWEET_FIELD,
            },
            auth=True,
        )

        users = [User(data, http_client=self) for data in res["data"]]
        for user in users:
            self.user_cache[user.id] = user
        return users

    def fetch_user_by_username(self, username: str) -> Optional[User]:
        if username.startswith("@"):
            username = username.replace("@", "", 1)

        data = self.request(
            "GET",
            "2",
            f"/users/by/username/{username}",
            params={
                "expansions": PINNED_TWEET_EXPANSION,
                "user.fields": USER_FIELD,
                "tweet.fields": TWEET_FIELD,
            },
            auth=True,
        )
        user = User(data, http_client=self)
        self.user_cache[user.id] = user
        return user

    def fetch_tweet(
        self, tweet_id: ID, *, organic_metrics: bool = False, promoted_metrics: bool = False
    ) -> Optional[Tweet]:
        field = TWEET_FIELD
        if organic_metrics and promoted_metrics:
            field = COMPLETE_TWEET_FIELD
        elif organic_metrics:
            field = TWEET_FIELD_WITH_ORGANIC_METRICS
        elif promoted_metrics:
            field = TWEET_FIELD_WITH_PROMOTED_METRICS

        res = self.request(
            "GET",
            "2",
            f"/tweets/{tweet_id}",
            params={
                "tweet.fields": field,
                "user.fields": USER_FIELD,
                "expansions": TWEET_EXPANSION,
                "media.fields": MEDIA_FIELD,
                "place.fields": PLACE_FIELD,
                "poll.fields": POLL_FIELD,
            },
            auth=True,
        )

        tweet = Tweet(res, http_client=self)
        self.tweet_cache[tweet.id] = tweet
        return tweet

    def fetch_space(self, space_id: str, *, space_host: bool) -> Space:
        res = self.request(
            "GET",
            "2",
            f"/spaces/{str(space_id)}",
            params={
                "expansions": SPACE_EXPANSION if not space_host else COMPLETE_SPACE_FIELD,
                "space.fields": SPACE_FIELD,
                "topic.fields": TOPIC_FIELD,
                "user.fields": USER_FIELD,
            },
        )
        return Space(res, http_client=self)

    def fetch_spaces_bytitle(
        self, title: str, state: SpaceState = SpaceState.live, *, space_host: bool
    ) -> Optional[List[Space]]:
        res = self.request(
            "GET",
            "2",
            "/spaces/search",
            params={
                "query": title,
                "state": state.value,
                "expansions": SPACE_EXPANSION if not space_host else COMPLETE_SPACE_FIELD,
                "space.fields": SPACE_FIELD,
                "topic.fields": TOPIC_FIELD,
            },
        )
        return [Space(data, http_client=self) for data in res]

    def fetch_list(self, id: ID) -> TwitterList:
        res = self.request(
            "GET",
            "2",
            f"/lists/{id}",
            auth=True,
            params={
                "expansions": LIST_EXPANSION,
                "list.fields": LIST_FIELD,
                "user.fields": USER_FIELD,
            },
        )
        return TwitterList(res, http_client=self)

    def handle_events(self, payload: Payload):
        if payload.get("direct_message_events"):
            self.event_parser.parse_direct_message_create(payload)

        elif payload.get("direct_message_indicate_typing_events"):
            self.event_parser.parse_direct_message_typing(payload)

        elif payload.get("direct_message_mark_read_events"):
            self.event_parser.parse_direct_message_read(payload)

        elif payload.get("favorite_events"):
            self.event_parser.parse_favorite_tweet(payload)

        elif payload.get("user_event"):
            self.event_parser.parse_revoke_event(payload)

        elif payload.get("follow_events"):
            self.event_parser.parse_user_action(payload, "follow_events")

        elif payload.get("block_events"):
            self.event_parser.parse_user_action(payload, "block_events")

        elif payload.get("mute_events"):
            self.event_parser.parse_user_action(payload, "mute_events")

        elif payload.get("tweet_create_events"):
            self.event_parser.parse_tweet_create(payload)

        elif payload.get("tweet_delete_events"):
            self.event_parser.parse_tweet_delete(payload)

    def fetch_direct_message(self, event_id: ID) -> Optional[DirectMessage]:
        try:
            event_id = str(event_id)
        except ValueError:
            raise ValueError("event_id must be an integer or a string of digits.")

        res = self.request("GET", "1.1", f"/direct_messages/events/show.json?id={event_id}", auth=True)

        message_create = res.get("event").get("message_create")
        recipient_id = int(message_create.get("target").get("recipient_id"))
        sender_id = int(message_create.get("sender_id"))

        recipient = self.user_cache.get(recipient_id)
        sender = self.user_cache.get(sender_id)
        if not recipient:
            recipient = self.fetch_user(recipient_id)

        if not sender:
            sender = self.fetch_user(sender_id)

        res["event"]["message_create"]["target"]["recipient"] = recipient
        res["event"]["message_create"]["target"]["sender"] = sender
        message = DirectMessage(res, http_client=self)
        self.message_cache[message.id] = message
        self.message_cache[recipient_id] = recipient
        self.message_cache[sender_id] = sender
        return message

    def fetch_welcome_message(self, welcome_message_id: ID) -> Optional[WelcomeMessage]:
        try:
            welcome_message_id = str(welcome_message_id)
        except ValueError:
            raise ValueError("welcome_message_id must be an integer or a string of digits.")

        res = self.request(
            "GET",
            "1.1",
            "/direct_messages/welcome_messages/show.json",
            params={"id": str(welcome_message_id)},
            auth=True,
        )

        data = res.get("welcome_message")
        message_data = data.get("message_data")
        id = data.get("id")
        timestamp = data.get("created_timestamp")
        text = message_data.get("text")
        return WelcomeMessage(text=text, id=id, timestamp=timestamp, http_client=self)

    def fetch_welcome_message_rule(self, welcome_message_rule_id) -> Optional[WelcomeMessageRule]:
        res = self.request(
            "GET",
            "1.1",
            "/direct_messages/welcome_messages/rules/show.json",
            params={"id": str(welcome_message_rule_id)},
            auth=True,
        )
        data = res.get("welcome_message_rule")
        id = data.get("id")
        timestamp = data.get("created_timestamp")
        welcome_message_id = data.get("welcome_message_id")
        return WelcomeMessageRule(id, welcome_message_id, timestamp, http_client=self)

    def fetch_job(self, id: ID) -> Optional[Job]:
        res = self.request(
            "GET",
            "2",
            f"/compliance/jobs/{id}",
        )

        return Job(res.get("data"), http_client=self)

    def fetch_jobs(self, type: JobType, status: Optional[JobStatus] = None) -> Optional[List[Job]]:
        res = self.request(
            "GET",
            "2",
            f"/compliance/jobs",
            params={"type": type.value, "status": status.value if isinstance(status, JobStatus) else status},
        )

        if not res:
            return []

        return [Job(data) for data in res["data"]]

    def search_geo(
        self,
        query: str,
        max_result: Optional[ID] = None,
        *,
        lat: Optional[int] = None,
        long: Optional[int] = None,
        ip: Optional[ID] = None,
        granularity: Granularity = Granularity.neighborhood,
    ) -> Optional[Geo]:
        if query:
            query = query.replace(" ", "%20")

        data = self.request(
            "GET",
            "1.1",
            "/geo/search.json",
            params={
                "query": query,
                "lat": lat,
                "long": long,
                "ip": ip,
                "granularity": granularity.value,
                "max_results": max_result,
            },
            auth=True,
        )

        return [Geo(data) for data in data.get("result").get("places")]

    def send_message(
        self,
        recipient_id: ID,
        text: str,
        *,
        file: Optional[File] = None,
        custom_profile: Optional[CustomProfile] = None,
        quick_reply: Optional[QuickReply] = None,
        cta: Optional[CTA] = None,
    ) -> Optional[NoReturn]:
        thread_session = self.thread_manager.generate_thread_session()
        data = {
            "event": {
                "type": "message_create",
                "message_create": {
                    "target": {"recipient_id": str(recipient_id)},
                    "message_data": {},
                },
            }
        }

        executor = self.thread_manager.create_new_executor(
            thread_name="post-tweet-file-request", session_id=thread_session
        )
        message_data = data["event"]["message_create"]["message_data"]
        message_data["text"] = str(text)

        if file:
            future = executor.submit(self.quick_upload, file)

        if custom_profile:
            message_data["custom_profile_id"] = str(custom_profile.id)

        if quick_reply:
            message_data["quick_reply"] = {
                "type": quick_reply.type,
                "options": quick_reply.raw_options,
            }

        if cta:
            message_data["ctas"] = cta.raw_buttons

        if file:
            file = future.result()
            message_data["attachment"] = {}
            message_data["attachment"]["type"] = "media"
            message_data["attachment"]["media"] = {}
            message_data["attachment"]["media"]["id"] = str(file.media_id)

        res = self.request(
            "POST",
            "1.1",
            "/direct_messages/events/new.json",
            json=data,
            auth=True,
        )

        message_create = res.get("event").get("message_create")
        user_id = message_create.get("target").get("recipient_id")
        user = self.fetch_user(user_id)
        res["event"]["message_create"]["target"]["recipient"] = user

        msg = DirectMessage(res, http_client=self)
        self.message_cache[msg.id] = msg
        return msg

    def post_tweet(
        self,
        text: str = None,
        *,
        file: Optional[File] = None,
        files: Optional[List[File]] = None,
        poll: Optional[Poll] = None,
        geo: Optional[Union[Geo, str]] = None,
        direct_message_deep_link: Optional[str] = None,
        reply_setting: Optional[Union[ReplySetting, str]] = None,
        quote_tweet: Optional[Union[Tweet, ID]] = None,
        reply_tweet: Optional[Union[Tweet, ID]] = None,
        exclude_reply_users: Optional[List[User, ID]] = None,
        media_tagged_users: Optional[List[User, ID]] = None,
        super_followers_only: bool = False,
    ) -> Optional[Tweet]:
        thread_session = self.thread_manager.generate_thread_session()
        executor = self.thread_manager.create_new_executor(thread_name="post-tweet-request", session_id=thread_session)

        payload = {}
        if text:
            payload["text"] = text

        if file:
            payload["media"] = {}
            payload["media"]["media_ids"] = []
            executor.submit(self.quick_upload, file)

        if files:
            if len(files) + 1 if file else len(files) > 4:
                raise BadRequests(message="Cannot upload more then 4 files!")

            payload["media"] = {}
            payload["media"]["media_ids"] = []
            for file in files:
                executor.submit(self.quick_upload, file)

        if poll:
            payload["poll"] = {}
            payload["poll"]["duration_minutes"] = int(poll.duration)
            payload["poll"]["options"] = [option.label for option in poll.options]

        if geo:
            payload["geo"] = {}
            payload["geo"]["place_id"] = geo.id if isinstance(geo, Geo) else geo

        if direct_message_deep_link:
            payload["direct_message_deep_link"] = direct_message_deep_link

        if reply_setting:
            payload["reply_settings"] = (
                reply_setting.value if isinstance(reply_setting, ReplySetting) else reply_setting
            )

        if reply_tweet:
            payload["reply"] = {}
            payload["reply"]["in_reply_to_tweet_id"] = (
                str(reply_tweet.id) if isinstance(reply_tweet, Tweet) else str(reply_tweet)
            )

        if quote_tweet:
            payload["quote_tweet_id"] = str(quote_tweet.id) if isinstance(quote_tweet, Tweet) else str(quote_tweet)

        if exclude_reply_users:
            ids = [str(user.id) if isinstance(user, User) else str(user) for user in exclude_reply_users]

            if "reply" in payload.keys():
                payload["reply"]["exclude_reply_user_ids"] = ids
            else:
                payload["reply"] = {}
                payload["reply"]["exclude_reply_user_ids"] = ids

        if media_tagged_users:
            if not payload.get("media"):
                raise pytwotException("Cannot tag users without any file!")
            payload["media"]["tagged_user_ids"] = [
                str(user.id) if isinstance(user, (User, ClientAccount)) else str(user) for user in media_tagged_users
            ]

        if super_followers_only:
            payload["for_super_followers_only"] = True

        executor.wait_for_futures()
        if file:
            payload["media"]["media_ids"].append(str(file.media_id))
        if files:
            for file in files:
                payload["media"]["media_ids"].append(str(file.media_id))

        res = self.request("POST", "2", "/tweets", json=payload, auth=True)
        return self.fetch_tweet(res["data"]["id"])

    def create_job(self, type: JobType, *, name: Optional[str] = None, resumable: bool = False) -> Optional[Job]:
        res = self.request(
            "POST", "2", "/compliance/jobs", json={"type": type.value, "name": name, "resumable": resumable}
        )

        return Job(res.get("data"))

    def create_list(self, name: str, *, description: str = "", private: bool = False) -> Optional[TwitterList]:
        res = self.request(
            "POST",
            "2",
            "/lists",
            auth=True,
            json={"name": name, "description": description, "private": private},
        )
        return TwitterList(res, http_client=self)

    def update_list(
        self,
        list_id: int,
        *,
        name: Optional[str] = None,
        description: str = "",
        private: Optional[bool] = None,
    ) -> Optional[RelationUpdate]:
        res = self.request(
            "PUT",
            "2",
            f"/lists/{list_id}",
            auth=True,
            json={"name": name, "description": description, "private": private},
        )
        return RelationUpdate(res)

    def create_custom_profile(self, name: str, file: File) -> Optional[CustomProfile]:
        file = self.quick_upload(file)
        data = {"custom_profile": {"name": name, "avatar": {"media": {"id": file.media_id}}}}

        res = self.request("POST", "1.1", "/custom_profiles/new.json", json=data, auth=True)
        data = res.get("custom_profile")

        return CustomProfile(
            data.get("name"),
            data.get("id"),
            data.get("created_timestamp"),
            data.get("avatar"),
        )

    def create_welcome_message(
        self,
        *,
        name: Optional[str] = None,
        text: Optional[str] = None,
        file: Optional[File] = None,
        quick_reply: Optional[QuickReply] = None,
        cta: Optional[CTA] = None,
    ) -> Optional[WelcomeMessage]:
        thread_session = self.thread_manager.generate_thread_session()
        executor = self.thread_manager.create_new_executor(
            thread_name="create-welcome-message-file-request", session_id=thread_session
        )
        data = {"welcome_message": {"message_data": {}}}
        message_data = data["welcome_message"]["message_data"]
        data["welcome_message"]["name"] = str(name)
        message_data["text"] = str(text)

        if file:
            file_future = executor.submit(self.quick_upload, file)

        if quick_reply:
            message_data["quick_reply"] = {
                "type": quick_reply.type,
                "options": quick_reply.raw_options,
            }

        if cta:
            message_data["ctas"] = cta.raw_buttons

        if file:
            message_data["attachment"] = {}
            message_data["attachment"]["type"] = "media"
            message_data["attachment"]["media"] = {}
            message_data["attachment"]["media"]["id"] = str(file_future.result().media_id)

        res = self.request(
            "POST",
            "1.1",
            "/direct_messages/welcome_messages/new.json",
            json=data,
            auth=True,
        )

        data = res.get("welcome_message")
        message_data = data.get("message_data")

        return WelcomeMessage(
            res.get("name"),
            text=message_data.get("text"),
            id=data.get("id"),
            timestamp=data.get("created_timestamp"),
            http_client=self,
        )

    def update_welcome_message(
        self,
        *,
        welcome_message_id: ID,
        text: Optional[str] = None,
        file: Optional[File] = None,
        quick_reply: Optional[QuickReply] = None,
        cta: Optional[CTA] = None,
    ):
        thread_session = self.thread_manager.generate_thread_session()
        executor = self.thread_manager.create_new_executor(
            thread_name="update-welcome-message-file-request", session_id=thread_session
        )
        data = {"message_data": {}}
        message_data = data["message_data"]
        message_data["text"] = str(text)

        if file:
            file_future = executor.submit(self.quick_upload, file)

        if quick_reply:
            message_data["quick_reply"] = {
                "type": quick_reply.type,
                "options": quick_reply.raw_options,
            }

        if cta:
            message_data["ctas"] = cta.raw_buttons

        if file:
            message_data["attachment"] = {}
            message_data["attachment"]["type"] = "media"
            message_data["attachment"]["media"] = {}
            message_data["attachment"]["media"]["id"] = str(file_future.result().media_id)

        res = self.request(
            "PUT",
            "1.1",
            "/direct_messages/welcome_messages/update.json",
            params={"id": str(welcome_message_id)},
            json=data,
            auth=True,
        )

        welcome_message = res.get("welcome_message")
        message_data = welcome_message.get("message_data")

        return WelcomeMessage(
            res.get("name"),
            text=message_data.get("text"),
            id=welcome_message.get("id"),
            timestamp=welcome_message.get("created_timestamp"),
            http_client=self,
        )



"""
The MIT License (MIT)

Copyright (c) 2021-present UnrealFar & TheGenocides
Copyright (c) 2023-present Sengolda

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""

from __future__ import annotations

import datetime
from typing import TYPE_CHECKING, Any, Dict, List, Optional, Union

from .attachments import Poll, Geo, File, Media
from .enums import ReplySetting
from .constants import (
    TWEET_EXPANSION,
    TWEET_FIELD,
    USER_FIELD,
    PINNED_TWEET_EXPANSION,
    MEDIA_FIELD,
    PLACE_FIELD,
    POLL_FIELD,
)
from .relations import RelationHide, RelationLike, RelationRetweet, RelationDelete
from .user import User
from .utils import time_parse_todt, convert
from .message import Message
from .paginations import UserPagination, TweetPagination
from .dataclass import (
    Embed,
    EmbedImage,
    PublicTweetMetrics,
    NonPublicTweetMetrics,
    OrganicTweetMetrics,
    PromotedTweetMetrics,
)

if TYPE_CHECKING:
    from .http import HTTPClient
    from .type import ID

__all__ = ("Tweet",)


class Tweet(Message):
    """Represents a tweet message from Twitter.
    A Tweet is any message posted to Twitter which may contain photos, videos, links, and text. This class inherits :class:`Message`,

    .. describe:: x == y

        Check if one tweet id is equal to another.


    .. describe:: x != y

        Check if one tweet id is not equal to another.


    .. describe:: str(x)

        Get the Tweet's text.


    .. versionadded:: 1.0.0
    """

    __slots__ = (
        "__original_payload",
        "_payload",
        "_includes",
        "tweet_metrics",
        "http_client",
        "deleted_timestamp",
        "_public_metrics",
        "_non_public_metrics",
        "_organic_metrics",
        "_promoted_metrics",
    )

    def __init__(
        self,
        data: Dict[str, Any],
        *,
        deleted_timestamp: Optional[int] = None,
        http_client: Optional[HTTPClient] = None,
    ) -> None:
        self.__original_payload = data
        self._payload = data.get("data") or data
        self._includes = self.__original_payload.get("includes")
        self._referenced_tweets = self._payload.get("referenced_tweets")
        self._entities = self._payload.get("entities")
        self.http_client = http_client
        self.deleted_timestamp = deleted_timestamp
        self._public_metrics = PublicTweetMetrics(
            **self._payload.get("public_metrics", None) or self.__original_payload.get("public_metrics")
        )
        self._non_public_metrics = self._payload.get("non_public_metrics", None) or self.__original_payload.get(
            "non_public_metrics"
        )
        self._organic_metrics = self._payload.get("organic_metrics", None) or self.__original_payload.get(
            "organic_metrics"
        )
        self._promoted_metrics = self._payload.get("promoted_metrics", None) or self.__original_payload.get(
            "promoted_metrics"
        )
        if self._non_public_metrics:
            non_public_metrics = self.http_client.payload_parser.parse_metric_data(self._non_public_metrics)
            self._non_public_metrics = NonPublicTweetMetrics(**non_public_metrics)

        if self._organic_metrics:
            organic_metrics = self.http_client.payload_parser.parse_metric_data(self._organic_metrics)
            self._organic_metrics = OrganicTweetMetrics(**organic_metrics)

        if self._promoted_metrics:
            promoted_metrics = self.http_client.payload_parser.parse_metric_data(self._promoted_metrics)
            self._promoted_metrics = PromotedTweetMetrics(**promoted_metrics)

        if self._entities and self._entities.get("urls"):
            data = []
            for url in self._entities["urls"]:
                url = self.http_client.payload_parser.parse_embed_data(url)
                data.append(url)
            self._embeds = data
        else:
            self._embeds = None

        super().__init__(self._payload.get("text"), self._payload.get("id"), 1)

    def __repr__(self) -> str:
        return "Tweet(text={0.text} id={0.id} author={0.author!r})".format(self)

    @property
    def author(self) -> Optional[User]:
        """Optional[:class:`User`]: Returns a user (object) who posted the tweet.

        .. versionadded: 1.0.0
        """
        if self._includes and self._includes.get("users"):
            return User(self._includes.get("users")[0], http_client=self.http_client)
        return None

    @property
    def possibly_sensitive(self) -> bool:
        """:class:`bool`: Returns True if the tweet is possible sensitive to some users, else False.

        .. versionadded: 1.0.0
        """
        return self._payload.get("possibly_sensitive")

    @property
    def sensitive(self) -> bool:
        """:class:`bool`: An alias to :meth:`Tweet.possibly_sensitive`.

        .. versionadded: 1.5.0
        """
        return self.possibly_sensitive

    @property
    def created_at(self) -> datetime.datetime:
        """:class:`datetime.datetime`: Returns a :class:`datetime.datetime` object when the tweet was created.

        .. versionadded: 1.0.0
        """
        if self._payload.get("timestamp", None):
            return datetime.datetime.fromtimestamp(int(self._payload.get("timestamp", None)) / 1000)
        return time_parse_todt(self._payload.get("created_at"))

    @property
    def deleted_at(self) -> Optional[datetime.datetime]:
        """Optional[:class:`datetime.datetime`]: Returns a :class:`datetime.datetime` object when the tweet was deleted. Returns None when the tweet is not deleted.

        .. note::
            This property can only returns :class:`datetime.datetime` object through a tweet object from `on_tweet_delete` event.

        .. versionadded: 1.5.0
        """
        if not self.deleted_timestamp:
            return None
        return datetime.datetime.fromtimestamp(self.deleted_timestamp / 1000)

    @property
    def source(self) -> str:
        """:class:`str`: Returns the source of the tweet. e.g if you post a tweet from a website, the source is gonna be 'Twitter Web App'

        .. versionadded: 1.0.0
        """
        return self._payload.get("source")

    @property
    def reply_setting(self) -> ReplySetting:
        """:class:`ReplySetting`: Returns a :class:`ReplySetting` object with the tweet's reply setting. If everyone can reply, this method return :class:`ReplySetting.everyone`.

        .. versionadded: 1.3.5
        """
        return ReplySetting(self._payload.get("reply_settings"))

    @property
    def raw_reply_setting(self) -> str:
        """:class:`str`: Returns the raw reply setting value. If everyone can replied, this method return 'Everyone'.

        .. versionadded: 1.0.0
        """
        return self._payload.get("reply_settings")

    @property
    def lang(self) -> str:
        """:class:`str`: Returns the tweet's lang, if its english it return en.

        .. versionadded: 1.0.0
        """
        return self._payload.get("lang")

    @property
    def conversation_id(self) -> Optional[int]:
        """Optional[:class:`int`]: All replies are bind to the original tweet, this property returns the tweet's id if the tweet is a reply tweet else it returns None.

        .. versionadded: 1.0.0
        """
        try:
            return int(self._payload.get("conversation_id"))
        except ValueError:
            return None

    @property
    def url(self) -> Optional[str]:
        """Optional[:class:`str`]: Get the tweet's url.

        .. versionadded:: 1.1.0

        .. versionchanged:: 1.5.0

            Returns None if the author is invalid or the tweet doesn't have id.
        """
        try:
            return f"https://twitter.com/{self.author.username}/status/{self.id}"
        except TypeError:
            return None

    @property
    def mentions(self) -> Optional[List[User]]:
        """Optional[List[:class:`User`]]: Returns a list of :class:`User` objects that were mentioned in the tweet or an empty list / `[]` if no users were mentioned.

        .. versionadded:: 1.1.3

        .. versionchanged:: 1.5.0

            Now returns a list of :class:`User` objects rather then a list of :class:`str` objects.
        """
        users = []
        for user in self._includes.get("users", {}):
            for mention in self._entities.get("mentions", {}):
                if user["id"] == mention["id"]:
                    users.append(User(user, http_client=self.http_client))
        return users

    @property
    def poll(self) -> Optional[Poll]:
        """:class:`Poll`: Returns a Poll object with the tweet's poll.

        .. versionadded:: 1.1.0
        """
        if self._includes:
            if self._includes.get("polls"):
                data = self._includes.get("polls")[0]
                poll = Poll(
                    duration=data.get("duration_minutes"),
                    id=data.get("id"),
                    voting_status=data.get("voting_status"),
                    end_date=data.get("end_datetime"),
                )
                for option in data.get("options"):
                    poll.add_option(**option)
                return poll
        return None

    @property
    def medias(self) -> Optional[List[Media]]:
        """Optional[List[:class:`Media`]]: Returns a list of media(s) in a tweet.

        .. versionadded:: 1.1.0
        """
        if self._includes and self._includes.get("media"):
            return [Media(img, http_client=self.http_client) for img in self._includes.get("media")]
        return None

    @property
    def reference_user(self) -> Optional[User]:
        """Optional[:class:`User`]: Returns the referenced user. This can means:

        The tweet is a retweet, which means the method returns the retweeted tweet's author.
        The tweet is a quote tweet(retweet with comments), which means the method returns the quoted tweet's author.
        The tweet is a reply tweet, which means the method returns the replied tweet's author.

        .. versionadded:: 1.5.0
        """
        if not self._includes or not self._includes.get("users") or not self._referenced_tweets:
            return None

        type = self._referenced_tweets[0].get("type", " ")
        for user in self._includes["users"]:
            if type == "replied_to" and user["id"] == self._payload.get("in_reply_to_user_id", 0):
                # If the tweet count as a reply tweet,
                # it would returns a user data that match the user's id with 'in_reply_to_user_id' data.
                return User(user, http_client=self.http_client)

            elif type == "quoted":
                # If the tweet count as a quote tweet,
                # it would returns a user data if the url contains the user's id. Every quote tweets have at least 1 url, the quoted tweet's url that contain the quoted tweet's author's id and the tweet id itself.
                for embed in self.embeds:
                    if (
                        embed.expanded_url.startswith("https://twitter.com/")
                        and embed.expanded_url.split("/")[3] == user["id"]
                    ):
                        return User(user, http_client=self.http_client)

            elif type == "retweeted":
                # If the tweet count as a retweet,
                # it would returns a user data if the user are mention with a specific format, that is: 'RT @Mention: {The retweeted tweet's content}'
                # The code only checks characters before the colon with the colon includes (i.e 'RT @Mention:').
                for mentioned_user in self.mentions:
                    if self.text.startswith(f"RT {mentioned_user.mention}:"):
                        return mentioned_user
        return None

    @property
    def reference_tweet(self) -> Optional[Tweet]:
        """Optional[:class:`Tweet`]: Returns the tweet's parent tweet or the  referenced tweet. This can mean the parent tweet of the requested tweet is:

        A retweeted tweet (The child Tweet is a Retweet),
        A quoted tweet (The child Tweet is a Retweet with comment, also known as Quoted Tweet),
        A replied tweet (The child Tweet is a reply tweet).

        .. versionadded:: 1.5.0
        """
        tweets = self._includes.get("tweets")
        if not self._includes or not tweets or not self._referenced_tweets:
            return None

        for tweet in tweets:
            if tweet["id"] == self._referenced_tweets[0]["id"]:
                self.http_client.payload_parser.insert_object_author(tweet, self.reference_user)
                return Tweet(tweet, http_client=self.http_client)
        return None

    @property
    def embeds(self) -> Optional[List[Embed]]:
        """List[:class:`Embed`]: Returns a list of Embedded urls from the tweet

        .. versionadded:: 1.1.3
        """
        for embed in self._embeds:
            if embed.get("images"):
                for index, image in enumerate(embed["images"]):
                    if isinstance(image, EmbedImage):
                        break
                    embed["images"][index] = EmbedImage(**image)
        return [Embed(**data) for data in self._embeds]

    @property
    def like_count(self) -> Optional[int]:
        """Optional[:class:`int`]: Returns the total of likes in the tweet.

        .. versionadded: 1.1.0
        """
        return convert(self._metrics.like_count, int)

    @property
    def retweet_count(self) -> Optional[int]:
        """Optional[:class:`int`]: Returns the total of retweetes in the tweet.

        .. versionadded: 1.1.0
        """
        return convert(self._metrics.retweet_count, int)

    @property
    def reply_count(self) -> Optional[int]:
        """Optional[:class:`int`]: Returns the total of replies in the tweet.

        .. versionadded: 1.1.0
        """
        return convert(self._metrics.reply_count, int)

    @property
    def quote_count(self) -> Optional[int]:
        """Optional[:class:`int`]: Returns the total of quotes in the tweet.

        .. versionadded: 1.1.0
        """
        return convert(self._metrics.quote_count, int)

    @property
    def non_public_metrics(self) -> Optional[OrganicTweetMetrics]:
        """Optional[:class:`OrganicTweetMetrics`]: The tweet's metrics that are not available for anyone to view on Twitter, such as `impressions_count` and `video view quartiles`.

        .. versionadded:: 1.5.0
        """
        return self._non_public_metrics

    @property
    def organic_metrics(self) -> Optional[OrganicTweetMetrics]:
        """Optional[:class:`OrganicTweetMetrics`]: The tweet's metrics in organic context (posted and viewed in a regular manner), such as `impression_count`, `user_profile_clicks` and `url_link_clicks`.

        .. versionadded:: 1.5.0
        """
        return self._organic_metrics

    @property
    def promoted_metrics(self) -> Optional[PromotedTweetMetrics]:
        """Optional[:class:`PromotedTweetMetrics`]: The tweet's metrics in promoted context (posted or viewed as part of an Ads campaign), such as `impression_count`, `user_profile_clicks` and `url_link_clicks`.

        .. versionadded:: 1.5.0
        """
        return self._promoted_metrics

    def like(self) -> Optional[RelationLike]:
        """Like the tweet.

        Returns
        ---------
        Optional[:class:`RelationLike`]
            This method returns a :class:`RelationLike` object.


        .. versionadded:: 1.2.0
        """
        my_id = self.http_client.access_token.partition("-")[0]
        res = self.http_client.request("POST", "2", f"/users/{my_id}/likes", json={"tweet_id": str(self.id)}, auth=True)
        return RelationLike(res)

    def unlike(self) -> Optional[RelationLike]:
        """Unlike the tweet.

        Returns
        ---------
        :class:`RelationLike`
            This method returns a :class:`RelationLike` object.


        .. versionadded:: 1.2.0
        """
        my_id = self.http_client.access_token.partition("-")[0]

        res = self.http_client.request("DELETE", "2", f"/users/{my_id}/likes/{self.id}", auth=True)

        return RelationLike(res)

    def retweet(self) -> RelationRetweet:
        """Retweet the tweet.

        Returns
        ---------
        :class:`RelationRetweet`
            This method returns a :class:`RelationRetweet` object.


        .. versionadded:: 1.2.0
        """
        my_id = self.http_client.access_token.partition("-")[0]
        res = self.http_client.request(
            "POST",
            "2",
            f"/users/{my_id}/retweets",
            json={"tweet_id": str(self.id)},
            auth=True,
        )

        return RelationRetweet(res)

    def unretweet(self) -> RelationRetweet:
        """Unretweet the tweet.

        Returns
        ---------
        :class:`RelationRetweet`
            This method returns a :class:`RelationRetweet` object.


        .. versionadded:: 1.2.0
        """
        my_id = self.http_client.access_token.partition("-")[0]

        res = self.http_client.request("DELETE", "2", f"/users/{my_id}/retweets/{self.id}", auth=True)

        return RelationRetweet(res)

    def delete(self) -> RelationDelete:
        """Delete the client's tweet.

        .. note::
            You can only delete the client's tweet.

        .. versionadded:: 1.2.0
        """
        res = self.http_client.request("DELETE", "2", f"/tweets/{self.id}", auth=True)

        try:
            self.http_client.tweet_cache.pop(self.id)
        except KeyError:
            pass

        return RelationDelete(res)

    def reply(
        self,
        text: str = None,
        *,
        file: Optional[File] = None,
        files: Optional[List[File]] = None,
        geo: Optional[Union[Geo, str]] = None,
        direct_message_deep_link: Optional[str] = None,
        reply_setting: Optional[Union[ReplySetting, str]] = None,
        exclude_reply_users: Optional[List[User, ID]] = None,
        media_tagged_users: Optional[List[User, ID]] = None,
    ) -> Optional[Tweet]:
        """Post a tweet to reply to the tweet present by the tweet's id. Returns a :class:`Tweet` object or :class:`Message` if the tweet is not found in the cache.

        .. note::
            Note that if the tweet is a retweet you cannot reply to that tweet, it might not raise an error but it will post the tweet has a normal tweet rather then a reply tweet and it ping the :class:`Tweet.author`.

        Parameters
        ------------
        text: :class:`str`
            The tweet's text, it will show up as the main text in a tweet.
        file: Optional[:class:`File`]
            Represents a single file attachment. It could be an image, gif, or video. It also have to be an instance of pytwot.File
        files: Optional[List[:class:`File`]]
            Represents multiple file attachments in a list. It could be an image, gif, or video. the item in the list must also be an instance of pytwot.File
        geo: Optional[Union[:class:`Geo`, :class:`str`]]
            The geo attachment, you can put an object that is an instance of :class:`Geo` or the place ID in a string.
        direct_message_deep_link: Optional[:class:`str`]
            The direct message deep link, It will showup as a CTA(call-to-action) with button attachment. Example of direct message deep link:
        reply_setting: Optional[Union[:class:`ReplySetting`, :class:`str`]]
            The reply setting that you can set to minimize users that can reply. If None is specified, the default is set to 'everyone' can reply.
        exclude_reply_users: Optional[List[:class:`User`]]
            A list of users or user ids to be excluded from the reply :class:`Tweet` thus removing a user from a thread, if you dont want to mention a reply with 3 mentions, You can use this argument and provide the user id you don't want to mention.
        media_tagged_users: Optional[List[:class:`User`]]
            A list of users or user ids being tagged in the Tweet with Media. If the user you're tagging doesn't have photo-tagging enabled, their names won't show up in the list of tagged users even though the Tweet is successfully created.

        Returns
        ---------
        Union[:class:`Tweet`, :class:`Message`]
            Returns a :class:`Tweet` object or :class:`Message` object if the tweet is not found in the cache.


        .. versionadded:: 1.2.5
        """
        return self.http_client.post_tweet(
            text,
            file=file,
            files=files,
            geo=geo,
            direct_message_deep_link=direct_message_deep_link,
            reply_setting=reply_setting,
            reply_tweet=self.id,
            exclude_reply_users=exclude_reply_users,
            media_tagged_users=media_tagged_users,
        )

    def hide(self) -> RelationHide:
        """Hide a reply tweet.

        Returns
        ---------
        :class:`RelationHide`
            This method returns a :class:`RelationHide` object.


        .. versionadded:: 1.2.5
        """
        res = self.http_client.request("PUT", "2", f"/tweets/{self.id}/hidden", json={"hidden": False}, auth=True)
        return RelationHide(res)

    def unhide(self) -> RelationHide:
        """Unhide a hide reply.

        Returns
        ---------
        :class:`RelationHide`
            This method returns a :class:`RelationHide` object.


        .. versionadded:: 1.2.5
        """
        res = self.http_client.request("PUT", "2", f"/tweets/{self.id}/hidden", json={"hidden": False}, auth=True)
        return RelationHide(res)

    def fetch_retweeters(self) -> Optional[UserPagination]:
        """Returns a pagination object with the users that retweeted the tweet.

        Returns
        ---------
        Optional[:class:`UserPagination`]
            This method returns a :class:`UserPagination` object.


        .. versionadded:: 1.1.3
        """
        res = self.http_client.request(
            "GET",
            "2",
            f"/tweets/{self.id}/retweeted_by",
            params={
                "expansions": PINNED_TWEET_EXPANSION,
                "user.fields": USER_FIELD,
                "tweet.fields": TWEET_FIELD,
            },
        )
        if not res:
            return []

        return UserPagination(
            res,
            endpoint_request=f"/tweets/{self.id}/retweeted_by",
            http_client=self.http_client,
            params={
                "expansions": PINNED_TWEET_EXPANSION,
                "user.fields": USER_FIELD,
                "tweet.fields": TWEET_FIELD,
            },
        )

    def fetch_likers(self) -> Optional[UserPagination]:
        """Returns a pagination object with the users that liked the tweet.

        Returns
        ---------
        Optional[:class:`UserPagination`]
            This method returns a :class:`UserPagination` object.


        .. versionadded:: 1.1.3
        """
        res = self.http_client.request(
            "GET",
            "2",
            f"/tweets/{self.id}/liking_users",
            params={
                "expansions": PINNED_TWEET_EXPANSION,
                "user.fields": USER_FIELD,
                "tweet.fields": TWEET_FIELD,
            },
        )

        if not res:
            return []

        return UserPagination(
            res,
            endpoint_request=f"/tweets/{self.id}/liking_users",
            http_client=self.http_client,
            params={
                "expansions": PINNED_TWEET_EXPANSION,
                "user.fields": USER_FIELD,
                "tweet.fields": TWEET_FIELD,
            },
        )

    def fetch_quoted_tweets(self) -> Optional[TweetPagination]:
        """Returns a pagination object for tweets that quoted the tweet

        Returns
        ---------
        Optional[:class:`TweetPagination`]
            This method returns :class:`TweetPagination` or an empty list if the tweet does not contain any quoted tweets.


        .. versionadded:: 1.5.0
        """
        params = {
            "expansions": TWEET_EXPANSION,
            "user.fields": USER_FIELD,
            "tweet.fields": TWEET_FIELD,
            "media.fields": MEDIA_FIELD,
            "place.fields": PLACE_FIELD,
            "poll.fields": POLL_FIELD,
            "max_results": 100,
        }

        res = self.http_client.request("GET", "2", f"/tweets/{self.id}/quote_tweets", params=params)

        if not res:
            return []

        return TweetPagination(
            res,
            endpoint_request=f"/tweets/{self.id}/quote_tweets",
            http_client=self.http_client,
            params=params,
        )



"""
The MIT License (MIT)

Copyright (c) 2021-present UnrealFar & TheGenocides
Copyright (c) 2023-present Sengolda

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""

from __future__ import annotations

from typing import Any, Dict, Optional, Tuple


class Hashtag:
    """Represents a hashtag in a message."""

    __slots__ = ("_payload", "_text", "_startpoint", "_endpoint")

    def __init__(self, data: Dict[str, Any]):
        self._payload = data
        self._text: Optional[str] = self._payload.get("text")
        self._startpoint, self._endpoint = self._payload.get("indices")

    @property
    def text(self) -> Optional[str]:
        """Optional[:class:`str`]: Returns the hashtag's text."""
        return self._text

    @property
    def points(self) -> Optional[Tuple]:
        """Optional[:class:`Tuple`]: Returns a tuple with the hashtag's startpoint and endpoint."""
        return self._startpoint, self._endpoint


class UserMention:
    """Represents a user mention in a message."""

    __slots__ = ("_payload", "_name", "_screen_name", "_id", "_startpoint", "_endpoint")

    def __init__(self, data: Dict[str, Any]):
        self._payload: Dict[str, Any] = data
        self._name: str = self._payload.get("name")
        self._screen_name: str = self._payload.get("screen_name")
        self._id: int = self._payload.get("id")
        self._startpoint, self._endpoint = self._payload.get("indices")

    @property
    def name(self) -> str:
        """:class:`str`: Returns the mention user's name."""
        return self._name

    @property
    def username(self) -> str:
        """:class:`str`: Returns the mention user's username."""
        return self._screen_name

    @property
    def id(self) -> int:
        """:class:`id`: Returns the mention user's id."""
        return int(self._id)

    @property
    def points(self) -> Optional[Tuple]:
        """Optional[:class:`Tuple`]: Returns a tuple with the mention's startpoint and endpoint."""
        return self._startpoint, self._endpoint


class Url:
    """Represents Url in a message."""

    __slots__ = (
        "_payload",
        "_url",
        "_display_url",
        "_expanded_url",
        "_startpoint",
        "_endpoint",
    )

    def __init__(self, data: Dict[str, Any]):
        self._payload: Dict[str, Any] = data
        self._url: str = self._payload.get("url")
        self._display_url: str = self._payload.get("display_url")
        self._expanded_url: str = self._payload.get("expanded_url")
        self._startpoint, self._endpoint = self._payload.get("indices")

    @property
    def url(self) -> str:
        """:class:`str`: Returns the message's url."""
        return self._url

    @property
    def display_url(self) -> str:
        """:class:`str`: Returns the message's display url"""
        return self._display_url

    @property
    def expanded_url(self) -> str:
        """:class:`str`: Returns the message's expanded url"""
        return self._expanded_url

    @property
    def points(self) -> Tuple:
        """Optional[:class:`Tuple`]: Returns a tuple with the url's startpoint and endpoint."""
        return self._startpoint, self._endpoint


class Symbol:
    """Represents a Symbol in a message."""

    __slots__ = ("_payload", "_text", "_startpoint", "_endpoint")

    def __init__(self, data=Optional[Dict[str, Any]]):
        self._payload: Dict[str, Any] = data
        self._text: str = self._payload.get("text")
        self._startpoint, self._endpoint = self._payload.get("indices")

    @property
    def text(self) -> str:
        """:class:`str`: Returns the symbol's text."""
        return self._text

    @property
    def points(self) -> Tuple:
        """Optional[:class:`Tuple`]: Returns a tuple with the url's startpoint and endpoint."""
        return self._startpoint, self._endpoint



from setuptools import setup
import re

short_description = "A Synchronous python API wrapper for twitter's api"
version = ""
with open("pytwot/__init__.py") as f:
    version = re.search(r'^__version__\s*=\s*[\'"]([^\'"]*)[\'"]', f.read(), re.MULTILINE).group(1)

if not version:
    raise RuntimeError("version is not set")

readme = ""
with open("README.md") as f:
    readme = f.read() or short_description

requirements = []
with open("requirements.txt") as f:
    requirements = f.read().splitlines()

extras_require = {
    "docs": [
        "sphinx>=4.0.2",
        "furo==2021.11.23",
        "sphinx_copybutton>=0.4.0",
    ],
    "events": ["Flask>=2.0.2"],
}

classifiers = [
    "License :: OSI Approved :: MIT License",
    "Programming Language :: Python :: 3.7",
    "Programming Language :: Python :: 3.8",
    "Programming Language :: Python :: 3.9",
    "Programming Language :: Python :: 3.10",
    "Operating System :: OS Independent",
    "Topic :: Internet",
    "Topic :: Utilities",
    "Development Status :: 5 - Production/Stable",
]

packages = [
    "pytwot",
    "pytwot.dataclass",
    "pytwot.threads"
]

# fmt: off

# fmt: on

setup(
    name="pytwot",
    author="Sengolda",
    maintainer="Sengolda",
    url="https://github.com/sengolda/pytwot",
    version=version,
    packages=packages,
    include_package_data=True,
    license="MIT",
    project_urls={
        "Documentation": "https://py-tweet.readthedocs.io/",
        "HomePage/Github": "https://github.com/sengolda/pytwot/",
        "Discord": "https://discord.gg/XHBhg6A4jJ",
    },
    description=short_description,
    long_description=readme,
    long_description_content_type="text/markdown",
    install_requires=requirements,
    extras_require=extras_require,
    keywords="twitter, tweet.py twitter.py",
    python_requires=">=3.7.0",
    classifiers=classifiers,
)
